@@ -54,10 +54,17 @@ static int cur_line = 1;
 static struct token head = { &head, &head, TOK_TEXT, 0, NULL, 0 };
 
 #define CONST_STR(str) { str, sizeof(str) - 1 }
-static const struct keyword {
+static const struct {
 	const char *str;
 	int len;
-} keywords[] = {
+} token_names[] = {
+	CONST_STR("text"),
+	CONST_STR("nl"),
+	CONST_STR("italic"),
+	CONST_STR("bold"),
+	CONST_STR("indent"),
+
+	// keywords
 	CONST_STR("h1"),
 	CONST_STR("h2"),
 	CONST_STR("li"),
@@ -68,23 +75,23 @@ static const struct keyword {
 	CONST_STR("endraw"),
 	CONST_STR("title")
 };
-#define NR_KEYWORDS (sizeof(keywords) / sizeof(struct keyword))
+#define NR_TOKEN_NAMES (sizeof(token_names) / sizeof(token_names[0]))
 #define BUG() die("BUG in %s\n", __FUNCTION__)
 
 #ifdef __GNUC__
-#define __NORETURN	__attribute__((__noreturn__))
+#define CMUS_NORETURN	__attribute__((__noreturn__))
 #else
-#define __NORETURN
+#define CMUS_NORETURN
 #endif
 
-static __NORETURN void quit(void)
+static CMUS_NORETURN void quit(void)
 {
 	if (tmp_file[0])
 		unlink(tmp_file);
 	exit(1);
 }
 
-static __NORETURN void die(const char *format, ...)
+static CMUS_NORETURN void die(const char *format, ...)
 {
 	va_list ap;
 
@@ -95,7 +102,7 @@ static __NORETURN void die(const char *format, ...)
 	quit();
 }
 
-static __NORETURN void syntax(int line, const char *format, ...)
+static CMUS_NORETURN void syntax(int line, const char *format, ...)
 {
 	va_list ap;
 
@@ -110,7 +117,7 @@ static inline const char *keyword_name(int type)
 {
 	if (type < TOK_H1 || type > TOK_TITLE)
 		die("BUG: no keyword name for type %d\n", type);
-	return keywords[type - TOK_H1].str;
+	return token_names[type].str;
 }
 
 static void *xmalloc(size_t size)
@@ -175,18 +182,18 @@ static int emit_keyword(const char *buf, int size)
 	int i, len;
 
 	for (len = 0; len < size; len++) {
-		if (!isalnum(buf[len]))
+		if (!isalnum((unsigned char)buf[len]))
 			break;
 	}
 
 	if (!len)
 		syntax(cur_line, "keyword expected\n");
 
-	for (i = 0; i < NR_KEYWORDS; i++) {
-		if (len != keywords[i].len)
+	for (i = TOK_H1; i < NR_TOKEN_NAMES; i++) {
+		if (len != token_names[i].len)
 			continue;
-		if (!strncmp(buf, keywords[i].str, len)) {
-			emit(TOK_H1 + i);
+		if (!strncmp(buf, token_names[i].str, len)) {
+			emit(i);
 			return len;
 		}
 	}
@@ -347,10 +354,12 @@ static struct token *get_indent(struct token *tok, int *ip)
 static struct token *check_line(struct token *tok, int *ip)
 {
 	struct token *start;
+	int tok_type;
 
 	start = tok = get_indent(tok, ip);
 
-	switch (tok->type) {
+	tok_type = tok->type;
+	switch (tok_type) {
 	case TOK_TEXT:
 	case TOK_BOLD:
 	case TOK_ITALIC:
@@ -390,7 +399,7 @@ static struct token *check_line(struct token *tok, int *ip)
 				return start;
 			default:
 				syntax(tok->line, "@%s can contain only text\n",
-						keyword_name(tok->type));
+						keyword_name(tok_type));
 			}
 			tok = tok->next;
 		}
