@@ -1,5 +1,5 @@
 /*
- * Copyright 2008-2011 Various Authors
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -16,8 +16,8 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
-#ifndef CACHE_H
-#define CACHE_H
+#ifndef CMUS_CACHE_H
+#define CMUS_CACHE_H
 
 #include "track_info.h"
 #include "locking.h"
@@ -29,8 +29,9 @@ extern pthread_mutex_t cache_mutex;
 
 int cache_init(void);
 int cache_close(void);
-struct track_info *cache_get_ti(const char *filename);
+struct track_info *cache_get_ti(const char *filename, int force);
 void cache_remove_ti(struct track_info *ti);
 struct track_info **cache_refresh(int *count, int force);
+struct track_info *lookup_cache_entry(const char *filename, unsigned int hash);
 
 #endif
