@@ -1,5 +1,5 @@
 /*
- * Copyright 2008-2011 Various Authors
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -16,8 +16,8 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
-#ifndef _CMUS_H
-#define _CMUS_H
+#ifndef CMUS_API_H
+#define CMUS_API_H
 
 #include "track_info.h"
 
@@ -72,7 +72,7 @@ enum file_type cmus_detect_ft(const char *name, char **ret);
  *
  * returns immediately, actual work is done in the worker thread.
  */
-void cmus_add(add_ti_cb, const char *name, enum file_type ft, int jt);
+void cmus_add(add_ti_cb, const char *name, enum file_type ft, int jt, int force);
 
 int cmus_save(for_each_ti_cb for_each_ti, const char *filename);
 int cmus_save_ext(for_each_ti_cb for_each_ti, const char *filename);
