@@ -1,5 +1,5 @@
 /*
- * Copyright 2008-2011 Various Authors
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -56,12 +56,6 @@
 #include <sys/wait.h>
 #include <dirent.h>
 
-#if defined(__sun__)
-#include <ncurses.h>
-#else
-#include <curses.h>
-#endif
-
 static struct history cmd_history;
 static char *cmd_history_filename;
 static char *history_search_text = NULL;
@@ -118,16 +112,16 @@ void view_add(int view, char *arg, int prepend)
 	switch (view) {
 	case TREE_VIEW:
 	case SORTED_VIEW:
-		cmus_add(lib_add_track, name, ft, JOB_TYPE_LIB);
+		cmus_add(lib_add_track, name, ft, JOB_TYPE_LIB, 0);
 		break;
 	case PLAYLIST_VIEW:
-		cmus_add(pl_add_track, name, ft, JOB_TYPE_PL);
+		cmus_add(pl_add_track, name, ft, JOB_TYPE_PL, 0);
 		break;
 	case QUEUE_VIEW:
 		if (prepend) {
-			cmus_add(play_queue_prepend, name, ft, JOB_TYPE_QUEUE);
+			cmus_add(play_queue_prepend, name, ft, JOB_TYPE_QUEUE, 0);
 		} else {
-			cmus_add(play_queue_append, name, ft, JOB_TYPE_QUEUE);
+			cmus_add(play_queue_append, name, ft, JOB_TYPE_QUEUE, 0);
 		}
 		break;
 	default:
@@ -165,7 +159,7 @@ void view_load(int view, char *arg)
 		editable_lock();
 		editable_clear(&lib_editable);
 		editable_unlock();
-		cmus_add(lib_add_track, name, FILE_TYPE_PL, JOB_TYPE_LIB);
+		cmus_add(lib_add_track, name, FILE_TYPE_PL, JOB_TYPE_LIB, 0);
 		free(lib_filename);
 		lib_filename = name;
 		break;
@@ -174,7 +168,7 @@ void view_load(int view, char *arg)
 		editable_lock();
 		editable_clear(&pl_editable);
 		editable_unlock();
-		cmus_add(pl_add_track, name, FILE_TYPE_PL, JOB_TYPE_PL);
+		cmus_add(pl_add_track, name, FILE_TYPE_PL, JOB_TYPE_PL, 0);
 		free(pl_filename);
 		pl_filename = name;
 		break;
@@ -353,6 +347,27 @@ static int flag_to_view(int flag)
 	}
 }
 
+struct window *current_win(void)
+{
+	switch (cur_view) {
+	case TREE_VIEW:
+		return lib_cur_win;
+	case SORTED_VIEW:
+		return lib_editable.win;
+	case PLAYLIST_VIEW:
+		return pl_editable.win;
+	case QUEUE_VIEW:
+		return pq_editable.win;
+	case BROWSER_VIEW:
+		return browser_win;
+	case HELP_VIEW:
+		return help_win;
+	case FILTERS_VIEW:
+	default:
+		return filters_win;
+	}
+}
+
 static void cmd_add(char *arg)
 {
 	int flag = parse_flags((const char **)&arg, "lpqQ");
@@ -425,6 +440,14 @@ static void cmd_set(char *arg)
 	if (value) {
 		option_set(arg, value);
 		help_win->changed = 1;
+		if (cur_view == TREE_VIEW) {
+			lib_track_win->changed = 1;
+			lib_tree_win->changed = 1;
+		} else {
+			current_win()->changed = 1;
+		}
+		update_titleline();
+		update_statusline();
 	} else {
 		struct cmus_opt *opt;
 		char buf[OPTION_MAX_SIZE];
@@ -455,6 +478,14 @@ static void cmd_toggle(char *arg)
 	}
 	opt->toggle(opt->id);
 	help_win->changed = 1;
+	if (cur_view == TREE_VIEW) {
+		lib_track_win->changed = 1;
+		lib_tree_win->changed = 1;
+	} else {
+		current_win()->changed = 1;
+	}
+	update_titleline();
+	update_statusline();
 }
 
 static int get_number(char *str, char **end)
@@ -477,6 +508,7 @@ static void cmd_seek(char *arg)
 	switch (*arg) {
 	case '-':
 		sign = -1;
+		/* fallthrough */
 	case '+':
 		relative = 1;
 		arg++;
@@ -509,8 +541,10 @@ inside:
 		switch (tolower((unsigned char)*arg)) {
 		case 'h':
 			seek *= 60;
+			/* fallthrough */
 		case 'm':
 			seek *= 60;
+			/* fallthrough */
 		case 's':
 			arg++;
 			break;
@@ -670,6 +704,8 @@ static void cmd_bind(char *arg)
 		goto err;
 
 	key_bind(arg, key, func, flag == 'f');
+	if (cur_view == HELP_VIEW)
+		window_changed(help_win);
 	return;
 err:
 	error_msg("expecting 3 arguments (context, key and function)\n");
@@ -864,7 +900,7 @@ static char *parse_one(const char **strp)
 	char *ret = NULL;
 
 	while (1) {
-		char *part;
+		char *part = NULL;
 		int c = *str;
 
 		if (!c || c == ' ')
@@ -888,12 +924,13 @@ static char *parse_one(const char **strp)
 		}
 
 		if (ret == NULL) {
-			ret = part;
+			ret = xstrdup(part);
 		} else {
 			char *tmp = xstrjoin(ret, part);
 			free(ret);
 			ret = tmp;
 		}
+		free(part);
 	}
 	*strp = str;
 	return ret;
@@ -904,7 +941,7 @@ error:
 	return NULL;
 }
 
-static char **parse_cmd(const char *cmd, int *args_idx, int *ac)
+char **parse_cmd(const char *cmd, int *args_idx, int *ac)
 {
 	char **av = NULL;
 	int nr = 0;
@@ -987,16 +1024,16 @@ static void cmd_run(char *arg)
 	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
-		__tree_for_each_sel(add_ti, NULL, 0);
+		_tree_for_each_sel(add_ti, &sel, 0);
 		break;
 	case SORTED_VIEW:
-		__editable_for_each_sel(&lib_editable, add_ti, &sel, 0);
+		_editable_for_each_sel(&lib_editable, add_ti, &sel, 0);
 		break;
 	case PLAYLIST_VIEW:
-		__editable_for_each_sel(&pl_editable, add_ti, &sel, 0);
+		_editable_for_each_sel(&pl_editable, add_ti, &sel, 0);
 		break;
 	case QUEUE_VIEW:
-		__editable_for_each_sel(&pq_editable, add_ti, &sel, 0);
+		_editable_for_each_sel(&pq_editable, add_ti, &sel, 0);
 		break;
 	}
 	editable_unlock();
@@ -1125,16 +1162,16 @@ static void cmd_echo(char *arg)
 	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
-		__tree_for_each_sel(get_one_ti, &sel_ti, 0);
+		_tree_for_each_sel(get_one_ti, &sel_ti, 0);
 		break;
 	case SORTED_VIEW:
-		__editable_for_each_sel(&lib_editable, get_one_ti, &sel_ti, 0);
+		_editable_for_each_sel(&lib_editable, get_one_ti, &sel_ti, 0);
 		break;
 	case PLAYLIST_VIEW:
-		__editable_for_each_sel(&pl_editable, get_one_ti, &sel_ti, 0);
+		_editable_for_each_sel(&pl_editable, get_one_ti, &sel_ti, 0);
 		break;
 	case QUEUE_VIEW:
-		__editable_for_each_sel(&pq_editable, get_one_ti, &sel_ti, 0);
+		_editable_for_each_sel(&pq_editable, get_one_ti, &sel_ti, 0);
 		break;
 	}
 	editable_unlock();
@@ -1295,7 +1332,11 @@ static void cmd_p_play(char *arg)
 
 static void cmd_p_prev(char *arg)
 {
-	cmus_prev();
+	if (rewind_offset < 0 || player_info.pos < rewind_offset) {
+		cmus_prev();
+	} else {
+		player_play();
+	}
 }
 
 static void cmd_p_stop(char *arg)
@@ -1303,6 +1344,14 @@ static void cmd_p_stop(char *arg)
 	player_stop();
 }
 
+static void cmd_pwd(char *arg)
+{
+	char buf[4096];
+	if (getcwd(buf, sizeof buf)) {
+		info_msg("%s", buf);
+	}
+}
+
 static void cmd_search_next(char *arg)
 {
 	if (search_str) {
@@ -1365,12 +1414,13 @@ static void add_from_browser(add_ti_cb add, int job_type)
 
 		if (ends_with(sel, "/../") || ends_with(sel, "/..")) {
 			info_msg("For convenience, you can not add \"..\" directory from the browser view");
+			free(sel);
 			return;
 		}
 
 		ft = cmus_detect_ft(sel, &ret);
 		if (ft != FILE_TYPE_INVALID) {
-			cmus_add(add, ret, ft, job_type);
+			cmus_add(add, ret, ft, job_type, 0);
 			window_down(browser_win, 1);
 		}
 		free(ret);
@@ -1442,17 +1492,31 @@ static void cmd_win_add_q(char *arg)
 static void cmd_win_activate(char *arg)
 {
 	struct track_info *info = NULL;
+	struct shuffle_track *previous = NULL, *next = NULL;
+	struct rb_root *shuffle_root = NULL;
+
+	if (cur_view == TREE_VIEW || cur_view == SORTED_VIEW) {
+		if (lib_cur_track)
+			previous = &lib_cur_track->shuffle_track;
+		shuffle_root = &lib_shuffle_root;
+	} else if (cur_view == PLAYLIST_VIEW) {
+		previous = (struct shuffle_track *)pl_cur_track;
+		shuffle_root = &pl_shuffle_root;
+	}
 
 	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
-		info = tree_set_selected();
+		info = tree_activate_selected();
+		next = &lib_cur_track->shuffle_track;
 		break;
 	case SORTED_VIEW:
-		info = sorted_set_selected();
+		info = sorted_activate_selected();
+		next = &lib_cur_track->shuffle_track;
 		break;
 	case PLAYLIST_VIEW:
-		info = pl_set_selected();
+		info = pl_activate_selected();
+		next = (struct shuffle_track *)pl_cur_track;
 		break;
 	case QUEUE_VIEW:
 		break;
@@ -1469,6 +1533,8 @@ static void cmd_win_activate(char *arg)
 	editable_unlock();
 
 	if (info) {
+		if (shuffle)
+			shuffle_insert(shuffle_root, previous, next);
 		/* update lib/pl mode */
 		if (cur_view < 2)
 			play_library = 1;
@@ -1563,7 +1629,7 @@ static void cmd_win_sel_cur(char *arg)
 	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
-		tree_sel_current();
+		tree_sel_current(auto_expand_albums_selcur);
 		break;
 	case SORTED_VIEW:
 		sorted_sel_current();
@@ -1617,25 +1683,18 @@ static void cmd_win_toggle(char *arg)
 	}
 }
 
-static struct window *current_win(void)
+static void cmd_win_scroll_down(char *arg)
 {
-	switch (cur_view) {
-	case TREE_VIEW:
-		return lib_cur_win;
-	case SORTED_VIEW:
-		return lib_editable.win;
-	case PLAYLIST_VIEW:
-		return pl_editable.win;
-	case QUEUE_VIEW:
-		return pq_editable.win;
-	case BROWSER_VIEW:
-		return browser_win;
-	case HELP_VIEW:
-		return help_win;
-	case FILTERS_VIEW:
-	default:
-		return filters_win;
-	}
+	editable_lock();
+	window_scroll_down(current_win());
+	editable_unlock();
+}
+
+static void cmd_win_scroll_up(char *arg)
+{
+	editable_lock();
+	window_scroll_up(current_win());
+	editable_unlock();
 }
 
 static void cmd_win_bottom(char *arg)
@@ -1647,8 +1706,18 @@ static void cmd_win_bottom(char *arg)
 
 static void cmd_win_down(char *arg)
 {
+	unsigned num_rows = 1;
+	char *end;
+
+	if (arg) {
+		if ((num_rows = get_number(arg, &end)) == 0 || *end) {
+			error_msg("invalid argument\n");
+			return;
+		}
+	}
+
 	editable_lock();
-	window_down(current_win(), 1);
+	window_down(current_win(), num_rows);
 	editable_unlock();
 }
 
@@ -1675,6 +1744,27 @@ static void cmd_win_pg_up(char *arg)
 	editable_unlock();
 }
 
+static void cmd_win_pg_top(char *arg)
+{
+	editable_lock();
+	window_page_top(current_win());
+	editable_unlock();
+}
+
+static void cmd_win_pg_bottom(char *arg)
+{
+	editable_lock();
+	window_page_bottom(current_win());
+	editable_unlock();
+}
+
+static void cmd_win_pg_middle(char *arg)
+{
+	editable_lock();
+	window_page_middle(current_win());
+	editable_unlock();
+}
+
 static void cmd_win_update_cache(char *arg)
 {
 	struct track_info_selection sel = { .tis = NULL };
@@ -1701,8 +1791,18 @@ static void cmd_win_top(char *arg)
 
 static void cmd_win_up(char *arg)
 {
+	unsigned num_rows = 1;
+	char *end;
+
+	if (arg) {
+		if ((num_rows = get_number(arg, &end)) == 0 || *end) {
+			error_msg("invalid argument\n");
+			return;
+		}
+	}
+
 	editable_lock();
-	window_up(current_win(), 1);
+	window_up(current_win(), num_rows);
 	editable_unlock();
 }
 
@@ -1713,6 +1813,12 @@ static void cmd_win_update(char *arg)
 	case SORTED_VIEW:
 		cmus_update_lib();
 		break;
+	case PLAYLIST_VIEW:
+		editable_lock();
+		editable_clear(&pl_editable);
+		editable_unlock();
+		cmus_add(pl_add_track, pl_filename, FILE_TYPE_PL, JOB_TYPE_PL, 0);
+		break;
 	case BROWSER_VIEW:
 		browser_reload();
 		break;
@@ -1864,7 +1970,7 @@ static void cmd_lqueue(char *arg)
 		struct album_list *a = container_of(item, struct album_list, node);
 		struct tree_track *t;
 		struct rb_node *tmp;
- 
+
 		rb_for_each_entry(t, tmp, &a->album->track_root, tree_node)
 			play_queue_append(tree_track_info(t));
 		free(a);
@@ -2475,6 +2581,7 @@ struct command commands[] = {
 	{ "player-stop",	cmd_p_stop,	0, 0, NULL,		  0, 0 },
 	{ "prev-view",		cmd_prev_view,	0, 0, NULL,		  0, 0 },
 	{ "push",		cmd_push,	1,-1, expand_commands,	  0, 0 },
+	{ "pwd",		cmd_pwd,	0, 0, NULL,		  0, 0 },
 	{ "quit",		cmd_quit,	0, 1, NULL,		  0, 0 },
 	{ "refresh",		cmd_refresh,	0, 0, NULL,		  0, 0 },
 	{ "run",		cmd_run,	1,-1, expand_program_paths, 0, CMD_UNSAFE },
@@ -2501,17 +2608,22 @@ struct command commands[] = {
 	{ "win-add-Q",		cmd_win_add_Q,	0, 0, NULL,		  0, 0 },
 	{ "win-add-q",		cmd_win_add_q,	0, 0, NULL,		  0, 0 },
 	{ "win-bottom",		cmd_win_bottom,	0, 0, NULL,		  0, 0 },
-	{ "win-down",		cmd_win_down,	0, 0, NULL,		  0, 0 },
+	{ "win-down",		cmd_win_down,	0, 1, NULL,		  0, 0 },
 	{ "win-mv-after",	cmd_win_mv_after,0, 0, NULL,		  0, 0 },
 	{ "win-mv-before",	cmd_win_mv_before,0, 0, NULL,		  0, 0 },
 	{ "win-next",		cmd_win_next,	0, 0, NULL,		  0, 0 },
+	{ "win-page-bottom",	cmd_win_pg_bottom,0, 0, NULL,		  0, 0 },
 	{ "win-page-down",	cmd_win_pg_down,0, 0, NULL,		  0, 0 },
+	{ "win-page-middle",	cmd_win_pg_middle,0, 0, NULL,		  0, 0 },
+	{ "win-page-top",	cmd_win_pg_top,	0, 0, NULL,		  0, 0 },
 	{ "win-page-up",	cmd_win_pg_up,	0, 0, NULL,		  0, 0 },
 	{ "win-remove",		cmd_win_remove,	0, 0, NULL,		  0, CMD_UNSAFE },
+	{ "win-scroll-down",	cmd_win_scroll_down,0, 0, NULL,		  0, 0 },
+	{ "win-scroll-up",	cmd_win_scroll_up,0, 0, NULL,		  0, 0 },
 	{ "win-sel-cur",	cmd_win_sel_cur,0, 0, NULL,		  0, 0 },
 	{ "win-toggle",		cmd_win_toggle,	0, 0, NULL,		  0, 0 },
 	{ "win-top",		cmd_win_top,	0, 0, NULL,		  0, 0 },
-	{ "win-up",		cmd_win_up,	0, 0, NULL,		  0, 0 },
+	{ "win-up",		cmd_win_up,	0, 1, NULL,		  0, 0 },
 	{ "win-update",		cmd_win_update,	0, 0, NULL,		  0, 0 },
 	{ "win-update-cache",	cmd_win_update_cache,0, 1, NULL,	  0, 0 },
 	{ "wq",			cmd_quit,	0, 1, NULL,		  0, 0 },
@@ -2761,9 +2873,11 @@ void run_parsed_command(char *cmd, char *arg)
 				error_msg("too many arguments\n");
 				break;
 			}
-			if (run_only_safe_commands && c->flags & CMD_UNSAFE) {
-				d_print("trying to execute unsafe command over net\n");
-				break;
+			if (run_only_safe_commands && (c->flags & CMD_UNSAFE)) {
+				if (c->func != cmd_save || !is_stdout_filename(arg)) {
+					d_print("trying to execute unsafe command over net\n");
+					break;
+				}
 			}
 			c->func(arg);
 			break;
@@ -2944,6 +3058,32 @@ void command_mode_key(int key)
 	reset_history_search();
 }
 
+void command_mode_mouse(MEVENT *event)
+{
+	if ((event->bstate & BUTTON1_PRESSED) || (event->bstate & BUTTON3_PRESSED)) {
+		if (event->y <= window_get_nr_rows(current_win()) + 2) {
+			if (cmdline.blen) {
+				history_add_line(&cmd_history, cmdline.line);
+				cmdline_clear();
+			}
+			input_mode = NORMAL_MODE;
+			normal_mode_mouse(event);
+			return;
+		}
+		if (event->x == 0)
+			return;
+		int i = event->x > cmdline.clen ? cmdline.clen : event->x - 1;
+		while (i < cmdline.cpos)
+			cmdline_move_left();
+		while (i > cmdline.cpos)
+			cmdline_move_right();
+	} else if (event->bstate & BUTTON4_PRESSED) {
+		command_mode_key(KEY_UP);
+	} else if (event->bstate & BUTTON5_PRESSED) {
+		command_mode_key(KEY_DOWN);
+	}
+}
+
 void commands_init(void)
 {
 	cmd_history_filename = xstrjoin(cmus_config_dir, "/command-history");
