@@ -1,5 +1,19 @@
 /*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2006 Timo Hirvonen
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "editable.h"
@@ -35,11 +49,12 @@ void editable_init(struct editable *e, void (*free_track)(struct list_head *item
 	struct iter iter;
 
 	list_init(&e->head);
+	e->tree_root = RB_ROOT;
 	e->nr_tracks = 0;
 	e->nr_marked = 0;
 	e->total_time = 0;
-	e->sort_keys = xnew(const char *, 1);
-	e->sort_keys[0] = NULL;
+	e->sort_keys = xnew(sort_key_t, 1);
+	e->sort_keys[0] = SORT_INVALID;
 	e->sort_str[0] = 0;
 	e->free_track = free_track;
 
@@ -52,15 +67,25 @@ void editable_init(struct editable *e, void (*free_track)(struct list_head *item
 	e->searchable = searchable_new(e->win, &iter, &simple_search_ops);
 }
 
-void editable_add(struct editable *e, struct simple_track *track)
+static void do_editable_add(struct editable *e, struct simple_track *track, int tiebreak)
 {
-	sorted_list_add_track(&e->head, track, e->sort_keys);
+	sorted_list_add_track(&e->head, &e->tree_root, track, e->sort_keys, tiebreak);
 	e->nr_tracks++;
 	if (track->info->duration != -1)
 		e->total_time += track->info->duration;
 	window_changed(e->win);
 }
 
+void editable_add(struct editable *e, struct simple_track *track)
+{
+	do_editable_add(e, track, +1);
+}
+
+void editable_add_before(struct editable *e, struct simple_track *track)
+{
+	do_editable_add(e, track, -1);
+}
+
 void editable_remove_track(struct editable *e, struct simple_track *track)
 {
 	struct track_info *ti = track->info;
@@ -74,8 +99,7 @@ void editable_remove_track(struct editable *e, struct simple_track *track)
 	if (ti->duration != -1)
 		e->total_time -= ti->duration;
 
-	list_del(&track->node);
-
+	sorted_list_remove_track(&e->head, &e->tree_root, track);
 	e->free_track(&track->node);
 }
 
@@ -101,50 +125,21 @@ void editable_remove_sel(struct editable *e)
 	}
 }
 
-static const char **sort_keys;
-
-static int list_cmp(const struct list_head *a_head, const struct list_head *b_head)
-{
-	const struct simple_track *a = to_simple_track(a_head);
-	const struct simple_track *b = to_simple_track(b_head);
-
-	return track_info_cmp(a->info, b->info, sort_keys);
-}
-
 void editable_sort(struct editable *e)
 {
-	sort_keys = e->sort_keys;
-	list_mergesort(&e->head, list_cmp);
+	if (e->nr_tracks <= 1)
+		return;
+	sorted_list_rebuild(&e->head, &e->tree_root, e->sort_keys);
+
 	window_changed(e->win);
 	window_goto_top(e->win);
 }
 
-static void keys_to_str(const char **keys, char *buf)
-{
-	int i, pos = 0;
-
-	for (i = 0; keys[i]; i++) {
-		const char *key = keys[i];
-		int len = strlen(key);
-
-		if (sizeof(buf) - pos - len - 2 < 0)
-			break;
-
-		memcpy(buf + pos, key, len);
-		pos += len;
-		buf[pos++] = ' ';
-	}
-	if (pos > 0)
-		pos--;
-	buf[pos] = 0;
-}
-
-void editable_set_sort_keys(struct editable *e, const char **keys)
+void editable_set_sort_keys(struct editable *e, sort_key_t *keys)
 {
 	free(e->sort_keys);
 	e->sort_keys = keys;
 	editable_sort(e);
-	keys_to_str(keys, e->sort_str);
 }
 
 void editable_toggle_mark(struct editable *e)
@@ -173,6 +168,18 @@ static void move_item(struct editable *e, struct list_head *head, struct list_he
 	list_add(item, head);
 }
 
+static void reset_tree(struct editable *e)
+{
+	struct simple_track *old, *first_track;
+
+	old = tree_node_to_simple_track(rb_first(&e->tree_root));
+	first_track = to_simple_track(e->head.next);
+	if (old != first_track) {
+		rb_replace_node(&old->tree_node, &first_track->tree_node, &e->tree_root);
+		RB_CLEAR_NODE(&old->tree_node);
+	}
+}
+
 static void move_sel(struct editable *e, struct list_head *after)
 {
 	struct simple_track *t;
@@ -193,7 +200,8 @@ static void move_sel(struct editable *e, struct list_head *after)
 	} else {
 		/* collect the selected track */
 		t = get_selected(e);
-		move_item(e, &tmp_head, &t->node);
+		if (t)
+			move_item(e, &tmp_head, &t->node);
 	}
 
 	/* put them back to the list after @after */
@@ -203,6 +211,7 @@ static void move_sel(struct editable *e, struct list_head *after)
 		list_add(item, after);
 		item = next;
 	}
+	reset_tree(e);
 
 	/* select top-most of the moved tracks */
 	editable_track_to_iter(e, to_simple_track(after->next), &iter);
@@ -259,10 +268,11 @@ void editable_move_after(struct editable *e)
 {
 	struct simple_track *sel;
 
-	if (e->nr_tracks <= 1)
+	if (e->nr_tracks <= 1 || e->sort_keys[0] != SORT_INVALID)
 		return;
 
-	if (e->sort_keys[0] == NULL && (sel = get_selected(e)))
+	sel = get_selected(e);
+	if (sel)
 		move_sel(e, find_insert_after_point(e, &sel->node));
 }
 
@@ -270,22 +280,31 @@ void editable_move_before(struct editable *e)
 {
 	struct simple_track *sel;
 
-	if (e->nr_tracks <= 1)
+	if (e->nr_tracks <= 1 || e->sort_keys[0] != SORT_INVALID)
 		return;
 
-	if (e->sort_keys[0] == NULL && (sel = get_selected(e)))
+	sel = get_selected(e);
+	if (sel)
 		move_sel(e, find_insert_before_point(e, &sel->node));
 }
 
 void editable_clear(struct editable *e)
 {
-	struct list_head *item, *next;
+	struct list_head *item, *tmp;
 
-	item = e->head.next;
-	while (item != &e->head) {
-		next = item->next;
+	list_for_each_safe(item, tmp, &e->head)
 		editable_remove_track(e, to_simple_track(item));
-		item = next;
+}
+
+void editable_remove_matching_tracks(struct editable *e,
+		int (*cb)(void *data, struct track_info *ti), void *data)
+{
+	struct list_head *item, *tmp;
+
+	list_for_each_safe(item, tmp, &e->head) {
+		struct simple_track *t = to_simple_track(item);
+		if (cb(data, t->info))
+			editable_remove_track(e, t);
 	}
 }
 
@@ -334,7 +353,7 @@ void editable_invert_marks(struct editable *e)
 	e->win->changed = 1;
 }
 
-int __editable_for_each_sel(struct editable *e, int (*cb)(void *data, struct track_info *ti),
+int _editable_for_each_sel(struct editable *e, int (*cb)(void *data, struct track_info *ti),
 		void *data, int reverse)
 {
 	int rc = 0;
@@ -356,8 +375,30 @@ int editable_for_each_sel(struct editable *e, int (*cb)(void *data, struct track
 {
 	int rc;
 
-	rc = __editable_for_each_sel(e, cb, data, reverse);
+	rc = _editable_for_each_sel(e, cb, data, reverse);
 	if (e->nr_marked == 0)
 		window_down(e->win, 1);
 	return rc;
 }
+
+void editable_update_track(struct editable *e, struct track_info *old, struct track_info *new)
+{
+	struct list_head *item, *tmp;
+	int changed = 0;
+
+	list_for_each_safe(item, tmp, &e->head) {
+		struct simple_track *track = to_simple_track(item);
+		if (track->info == old) {
+			if (new) {
+				track_info_unref(old);
+				track_info_ref(new);
+				track->info = new;
+			} else {
+				editable_remove_track(e, track);
+			}
+			changed = 1;
+		}
+	}
+	if (changed)
+		e->win->changed = changed;
+}
