@@ -1,6 +1,21 @@
 /*
- * help.c, heavily based on filters.c
- *  (c) 2006, <ft@bewatermyfriend.org>
+ * Copyright 2008-2013 Various Authors
+ * Copyright 2006 <ft@bewatermyfriend.org>
+ *
+ * heavily based on filters.c
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "help.h"
@@ -192,7 +207,7 @@ void help_select(void)
 {
 	struct iter sel;
 	struct help_entry *ent;
-	char buf[512];
+	char buf[OPTION_MAX_SIZE];
 
 	if (!window_get_sel(help_win, &sel))
 		return;
@@ -207,6 +222,12 @@ void help_select(void)
 		cmdline_set_text(buf);
 		enter_command_mode();
 		break;
+	case HE_UNBOUND:
+		snprintf(buf, sizeof(buf), "bind common <key> %s",
+				ent->command->name);
+		cmdline_set_text(buf);
+		enter_command_mode();
+		break;
 	case HE_OPTION:
 		snprintf(buf, sizeof(buf), "set %s=", ent->option->name);
 		ent->option->get(ent->option->id, buf + strlen(buf));
