@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,60 +11,30 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "http.h"
 #include "file.h"
 #include "debug.h"
 #include "xmalloc.h"
+#include "gbuf.h"
 
+#include <stdio.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/socket.h>
+#include <sys/select.h>
 #include <netinet/in.h>
 #include <netdb.h>
 #include <arpa/inet.h>
 #include <string.h>
 #include <errno.h>
 
-static void buf_ensure_space(char **bufp, int *sizep, int *posp, int len)
-{
-	int size = *sizep;
-	int pos = *posp;
-
-	if (size - pos < len) {
-		if (size == 0)
-			size = 128;
-		while (size - pos < len)
-			size *= 2;
-		*bufp = xrenew(char, *bufp, size);
-		*sizep = size;
-	}
-}
-
-static void buf_write(char **bufp, int *sizep, int *posp, const char *str)
-{
-	int len = strlen(str);
-
-	buf_ensure_space(bufp, sizep, posp, len);
-	memcpy(*bufp + *posp, str, len);
-	*posp += len;
-}
-
-static void buf_write_ch(char **bufp, int *sizep, int *posp, char ch)
-{
-	buf_ensure_space(bufp, sizep, posp, 1);
-	(*bufp)[*posp] = ch;
-	*posp += 1;
-}
-
 /*
  * @uri is http://[user[:pass]@]host[:port][/path][?query]
  *
@@ -76,6 +47,7 @@ int http_parse_uri(const char *uri, struct http_uri *u)
 	const char *str, *colon, *at, *slash, *host_start;
 
 	/* initialize all fields */
+	u->uri  = xstrdup(uri);
 	u->user = NULL;
 	u->pass = NULL;
 	u->host = NULL;
@@ -147,34 +119,61 @@ int http_parse_uri(const char *uri, struct http_uri *u)
 
 void http_free_uri(struct http_uri *u)
 {
+	free(u->uri);
 	free(u->user);
 	free(u->pass);
 	free(u->host);
 	free(u->path);
+
+	u->uri  = NULL;
+	u->user = NULL;
+	u->pass = NULL;
+	u->host = NULL;
+	u->path = NULL;
 }
 
-int http_open(const char *hostname, unsigned int port, int timeout_ms)
+int http_open(struct http_get *hg, int timeout_ms)
 {
-	struct hostent *hostent;
-	struct sockaddr_in addr;
+	const struct addrinfo hints = {
+		.ai_socktype = SOCK_STREAM
+	};
+	struct addrinfo *result;
+	union {
+		struct sockaddr sa;
+		struct sockaddr_storage sas;
+	} addr;
+	size_t addrlen;
 	struct timeval tv;
-	int sock, save, flags;
+	int save, flags, rc;
+	char port[16];
 
-	hostent = gethostbyname(hostname);
-	if (hostent == NULL)
+	char *proxy = getenv("http_proxy");
+	if (proxy) {
+		hg->proxy = xnew(struct http_uri, 1);
+		if (http_parse_uri(proxy, hg->proxy)) {
+			d_print("Failed to parse HTTP proxy URI '%s'\n", proxy);
+			return -1;
+		}
+	} else {
+		hg->proxy = NULL;
+	}
+
+	snprintf(port, sizeof(port), "%d", hg->proxy ? hg->proxy->port : hg->uri.port);
+	rc = getaddrinfo(hg->proxy ? hg->proxy->host : hg->uri.host, port, &hints, &result);
+	if (rc != 0) {
+		d_print("getaddrinfo: %s\n", gai_strerror(rc));
 		return -1;
-	if (hostent->h_length > sizeof(addr.sin_addr))
-		hostent->h_length = sizeof(addr.sin_addr);
-	addr.sin_family = AF_INET;
-	addr.sin_port = htons(port);
-	memcpy(&addr.sin_addr, hostent->h_addr_list[0], hostent->h_length);
+	}
+	memcpy(&addr.sa, result->ai_addr, result->ai_addrlen);
+	addrlen = result->ai_addrlen;
+	freeaddrinfo(result);
 
-	sock = socket(PF_INET, SOCK_STREAM, 0);
-	if (sock == -1)
+	hg->fd = socket(addr.sa.sa_family, SOCK_STREAM, 0);
+	if (hg->fd == -1)
 		return -1;
 
-	flags = fcntl(sock, F_GETFL);
-	if (fcntl(sock, F_SETFL, O_NONBLOCK) == -1)
+	flags = fcntl(hg->fd, F_GETFL);
+	if (fcntl(hg->fd, F_SETFL, O_NONBLOCK) == -1)
 		goto close_exit;
 
 	tv.tv_sec = timeout_ms / 1000;
@@ -182,8 +181,8 @@ int http_open(const char *hostname, unsigned int port, int timeout_ms)
 	while (1) {
 		fd_set wfds;
 
-		d_print("connecting. timeout=%ld s %ld us\n", tv.tv_sec, tv.tv_usec);
-		if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) == 0)
+		d_print("connecting. timeout=%lld s %lld us\n", (long long)tv.tv_sec, (long long)tv.tv_usec);
+		if (connect(hg->fd, &addr.sa, addrlen) == 0)
 			break;
 		if (errno == EISCONN)
 			break;
@@ -191,11 +190,9 @@ int http_open(const char *hostname, unsigned int port, int timeout_ms)
 			goto close_exit;
 
 		FD_ZERO(&wfds);
-		FD_SET(sock, &wfds);
+		FD_SET(hg->fd, &wfds);
 		while (1) {
-			int rc;
-
-			rc = select(sock + 1, NULL, &wfds, NULL, &tv);
+			rc = select(hg->fd + 1, NULL, &wfds, NULL, &tv);
 			if (rc == -1) {
 				if (errno != EINTR)
 					goto close_exit;
@@ -214,12 +211,12 @@ int http_open(const char *hostname, unsigned int port, int timeout_ms)
 	}
 
 	/* restore old flags */
-	if (fcntl(sock, F_SETFL, flags) == -1)
+	if (fcntl(hg->fd, F_SETFL, flags) == -1)
 		goto close_exit;
-	return sock;
+	return 0;
 close_exit:
 	save = errno;
-	close(sock);
+	close(hg->fd);
 	errno = save;
 	return -1;
 }
@@ -235,7 +232,7 @@ static int http_write(int fd, const char *buf, int count, int timeout_ms)
 		fd_set wfds;
 		int rc;
 
-		d_print("timeout=%ld s %ld us\n", tv.tv_sec, tv.tv_usec);
+		d_print("timeout=%lld s %lld us\n", (long long)tv.tv_sec, (long long)tv.tv_usec);
 
 		FD_ZERO(&wfds);
 		FD_SET(fd, &wfds);
@@ -292,13 +289,13 @@ static int read_timeout(int fd, int timeout_ms)
 }
 
 /* reads response, ignores fscking carriage returns */
-static int http_read_response(int fd, char **bufp, int *sizep, int *posp, int timeout_ms)
+static int http_read_response(int fd, struct gbuf *buf, int timeout_ms)
 {
+	char prev = 0;
+
 	if (read_timeout(fd, timeout_ms))
 		return -1;
 	while (1) {
-		char *buf = *bufp;
-		int pos = *posp;
 		int rc;
 		char ch;
 
@@ -311,29 +308,22 @@ static int http_read_response(int fd, char **bufp, int *sizep, int *posp, int ti
 		}
 		if (ch == '\r')
 			continue;
-		if (ch == '\n' && pos > 0 && buf[pos - 1] == '\n') {
-			buf_write_ch(bufp, sizep, posp, 0);
+		if (ch == '\n' && prev == '\n')
 			return 0;
-		}
-		buf_write_ch(bufp, sizep, posp, ch);
+		gbuf_add_ch(buf, ch);
+		prev = ch;
 	}
 }
 
-static int http_parse_response(const char *str, int *codep, char **reasonp, struct http_header **hp)
+static int http_parse_response(char *str, struct http_get *hg)
 {
 	/* str is 0 terminated buffer of lines
 	 * every line ends with '\n'
 	 * no carriage returns
 	 * no empty lines
 	 */
-	const char *end;
-	char *reason;
-	int code, i, count;
-	struct http_header *h;
-
-	*codep = -1;
-	*reasonp = NULL;
-	*hp = NULL;
+	GROWING_KEYVALS(h);
+	char *end;
 
 	if (strncmp(str, "HTTP/", 5) == 0) {
 		str += 5;
@@ -351,149 +341,114 @@ static int http_parse_response(const char *str, int *codep, char **reasonp, stru
 	while (*str == ' ')
 		str++;
 
-	code = 0;
-	for (i = 0; i < 3; i++) {
-		if (*str < '0' || *str > '9') {
-			return -2;
-		}
-		code *= 10;
-		code += *str - '0';
+	hg->code = 0;
+	while (*str >= '0' && *str <= '9') {
+		hg->code *= 10;
+		hg->code += *str - '0';
 		str++;
 	}
+	if (!hg->code)
+		return -2;
 	while (*str == ' ')
 		str++;
 
 	end = strchr(str, '\n');
-	reason = xstrndup(str, end - str);
+	hg->reason = xstrndup(str, end - str);
 	str = end + 1;
 
 	/* headers */
-	count = 4;
-	h = xnew(struct http_header, count);
-	i = 0;
 	while (*str) {
-		const char *ptr;
-
-		if (i == count - 1) {
-			count *= 2;
-			h = xrenew(struct http_header, h, count);
-		}
+		char *ptr;
 
 		end = strchr(str, '\n');
 		ptr = strchr(str, ':');
 		if (ptr == NULL || ptr > end) {
-			int j;
-
-			for (j = 0; j < i; j++) {
-				free(h[j].key);
-				free(h[j].val);
-			}
-			free(h);
-			free(reason);
-			*reasonp = NULL;
+			free(hg->reason);
+			hg->reason = NULL;
+			keyvals_terminate(&h);
+			keyvals_free(h.keyvals);
 			return -2;
 		}
-		h[i].key = xstrndup(str, ptr - str);
-		ptr++;
+
+		*ptr++ = 0;
 		while (*ptr == ' ')
 			ptr++;
-		h[i].val = xstrndup(ptr, end - ptr);
-		i++;
+
+		keyvals_add(&h, str, xstrndup(ptr, end - ptr));
 		str = end + 1;
 	}
-	h[i].key = NULL;
-	h[i].val = NULL;
-	*codep = code;
-	*reasonp = reason;
-	*hp = h;
+	keyvals_terminate(&h);
+	hg->headers = h.keyvals;
 	return 0;
 }
 
-int http_get(int fd, const char *path, struct http_header *headers,
-		int *codep, char **reasonp, struct http_header **ret_headersp,
-		int timeout_ms)
+int http_get(struct http_get *hg, struct keyval *headers, int timeout_ms)
 {
-	char *buf = NULL;
-	int size = 0;
-	int pos = 0;
+	GBUF(buf);
 	int i, rc, save;
 
-	buf_write(&buf, &size, &pos, "GET ");
-	buf_write(&buf, &size, &pos, path);
-	buf_write(&buf, &size, &pos, " HTTP/1.0\r\n");
+	gbuf_add_str(&buf, "GET ");
+	gbuf_add_str(&buf, hg->proxy ? hg->uri.uri : hg->uri.path);
+	gbuf_add_str(&buf, " HTTP/1.0\r\n");
 	for (i = 0; headers[i].key; i++) {
-		buf_write(&buf, &size, &pos, headers[i].key);
-		buf_write(&buf, &size, &pos, ": ");
-		buf_write(&buf, &size, &pos, headers[i].val);
-		buf_write(&buf, &size, &pos, "\r\n");
+		gbuf_add_str(&buf, headers[i].key);
+		gbuf_add_str(&buf, ": ");
+		gbuf_add_str(&buf, headers[i].val);
+		gbuf_add_str(&buf, "\r\n");
 	}
-	buf_write(&buf, &size, &pos, "\r\n");
-	
-	rc = http_write(fd, buf, pos, timeout_ms);
+	gbuf_add_str(&buf, "\r\n");
+
+	rc = http_write(hg->fd, buf.buffer, buf.len, timeout_ms);
 	if (rc)
 		goto out;
 
-	pos = 0;
-	rc = http_read_response(fd, &buf, &size, &pos, timeout_ms);
+	gbuf_clear(&buf);
+	rc = http_read_response(hg->fd, &buf, timeout_ms);
 	if (rc)
 		goto out;
 
-	rc = http_parse_response(buf, codep, reasonp, ret_headersp);
+	rc = http_parse_response(buf.buffer, hg);
 out:
 	save = errno;
-	free(buf);
+	gbuf_free(&buf);
 	errno = save;
 	return rc;
 }
 
-int http_read_body(int fd, char **bodyp, int timeout_ms)
+char *http_read_body(int fd, size_t *size, int timeout_ms)
 {
-	char *body = NULL;
-	int size = 0;
-	int pos = 0;
+	GBUF(buf);
 
-	*bodyp = NULL;
 	if (read_timeout(fd, timeout_ms))
-		return -1;
+		return NULL;
 	while (1) {
+		int count = 1023;
 		int rc;
 
-		buf_ensure_space(&body, &size, &pos, 256);
-		rc = read_all(fd, body + pos, size - pos);
+		gbuf_grow(&buf, count);
+		rc = read_all(fd, buf.buffer + buf.len, count);
 		if (rc == -1) {
-			free(body);
-			return -1;
+			gbuf_free(&buf);
+			return NULL;
 		}
+		buf.len += rc;
 		if (rc == 0) {
-			buf_ensure_space(&body, &size, &pos, 1);
-			body[pos] = 0;
-			*bodyp = body;
-			return 0;
+			*size = buf.len;
+			return gbuf_steal(&buf);
 		}
-		pos += rc;
 	}
 }
 
-const char *http_headers_get_value(const struct http_header *headers, const char *key)
+void http_get_free(struct http_get *hg)
 {
-	int i;
-
-	for (i = 0; headers[i].key; i++) {
-		if (strcasecmp(headers[i].key, key) == 0)
-			return headers[i].val;
+	http_free_uri(&hg->uri);
+	if (hg->proxy) {
+		http_free_uri(hg->proxy);
+		free(hg->proxy);
 	}
-	return NULL;
-}
-
-void http_headers_free(struct http_header *headers)
-{
-	int i;
-
-	for (i = 0; headers[i].key; i++) {
-		free(headers[i].key);
-		free(headers[i].val);
-	}
-	free(headers);
+	if (hg->headers)
+		keyvals_free(hg->headers);
+	free(hg->reason);
 }
 
 char *base64_encode(const char *str)
