@@ -1,19 +1,46 @@
 /*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2005 Timo Hirvonen
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "id3.h"
-#include "comment.h"
 #include "xmalloc.h"
-#include "utf8_encode.h"
+#include "convert.h"
 #include "uchar.h"
 #include "options.h"
 #include "debug.h"
+#include "utils.h"
+#include "file.h"
 
 #include <unistd.h>
-#include <inttypes.h>
+#include <stdint.h>
 #include <errno.h>
 #include <stdio.h>
+#include <string.h>
+#include <strings.h>
+#include <limits.h>
+
+enum {
+	ID3_ENCODING_ISO_8859_1 = 0x00,
+	ID3_ENCODING_UTF_16     = 0x01,
+	ID3_ENCODING_UTF_16_BE  = 0x02,
+	ID3_ENCODING_UTF_8      = 0x03,
+
+	ID3_ENCODING_MAX        = 0x03
+};
 
 /*
  * position:
@@ -207,85 +234,95 @@ static const char *genres[NR_GENRES] = {
 };
 /* }}} */
 
-#if 1
 #define id3_debug(...) d_print(__VA_ARGS__)
-#else
-#define id3_debug(...) do { } while (0)
-#endif
 
 const char * const id3_key_names[NUM_ID3_KEYS] = {
 	"artist",
 	"album",
 	"title",
 	"date",
+	"originaldate",
 	"genre",
 	"discnumber",
 	"tracknumber",
 	"albumartist",
 	"artistsort",
 	"albumartistsort",
+	"albumsort",
+	"compilation",
 	"replaygain_track_gain",
 	"replaygain_track_peak",
 	"replaygain_album_gain",
-	"replaygain_album_peak"
+	"replaygain_album_peak",
+	"composer",
+	"conductor",
+	"lyricist",
+	"remixer",
+	"label",
+	"publisher",
+	"subtitle",
+	"comment",
+	"musicbrainz_trackid",
+	"media",
 };
 
-static int utf16_is_special(const uchar uch)
+static int utf16_is_lsurrogate(uchar uch)
 {
-	if (UTF16_IS_HSURROGATE(uch) || UTF16_IS_LSURROGATE(uch) || UTF16_IS_BOM(uch))
-		return -1;
-	return 0;
+	return 0xdc00 <= uch && 0xdfff >= uch;
+}
+
+static int utf16_is_hsurrogate(uchar uch)
+{
+	return 0xd800 <= uch && 0xdbff >= uch;
+}
+
+static int utf16_is_bom(uchar uch)
+{
+	return uch == 0xfeff;
+}
+
+static int utf16_is_special(uchar uch)
+{
+	return utf16_is_hsurrogate(uch) || utf16_is_lsurrogate(uch) || utf16_is_bom(uch);
 }
 
 static char *utf16_to_utf8(const unsigned char *buf, int buf_size)
 {
 	char *out;
 	int i, idx;
+	int little_endian = 0;
+
+	if (buf_size < 2)
+		return NULL;
+
+	if (buf[0] == 0xff && buf[1] == 0xfe)
+		little_endian = 1;
 
 	out = xnew(char, (buf_size / 2) * 4 + 1);
 	i = idx = 0;
+
 	while (buf_size - i >= 2) {
 		uchar u;
 
-		u = buf[i] + (buf[i + 1] << 8);
+		if (little_endian)
+			u = buf[i] + (buf[i + 1] << 8);
+		else
+			u = buf[i + 1] + (buf[i] << 8);
+
 		if (u_is_unicode(u)) {
-			if (utf16_is_special(u) == 0)
+			if (!utf16_is_special(u))
 				u_set_char(out, &idx, u);
 		} else {
 			free(out);
 			return NULL;
 		}
+
 		if (u == 0)
 			return out;
+
 		i += 2;
 	}
-	u_set_char(out, &idx, 0);
-	return out;
-}
 
-static char *utf16be_to_utf8(const unsigned char *buf, int buf_size)
-{
-	char *out;
-	int i, idx;
-
-	out = xnew(char, (buf_size / 2) * 4 + 1);
-	i = 0;
-	idx = 0;
-	while (buf_size - i >= 2) {
-		uchar u;
-
-		u = buf[i + 1] + (buf[i] << 8);
-		if (u_is_unicode(u)) {
-		       if (utf16_is_special(u) == 0)
-				u_set_char(out, &idx, u);
-		} else {
-			free(out);
-			return NULL;
-		}
-		if (u == 0)
-			return out;
-		i += 2;
-	}
 	u_set_char(out, &idx, 0);
 	return out;
 }
@@ -337,6 +374,19 @@ static void get_u24(const unsigned char *buf, uint32_t *up)
 	*up = u;
 }
 
+static void get_i16(const unsigned char *buf, int16_t *ip)
+{
+	uint16_t b, u = 0;
+	int i;
+
+	for (i = 0; i < 2; i++) {
+		b = buf[i];
+		u <<= 8;
+		u |= b;
+	}
+	*ip = u;
+}
+
 static int v2_header_footer_parse(struct v2_header *header, const char *buf)
 {
 	const unsigned char *b = (const unsigned char *)buf;
@@ -414,7 +464,7 @@ static int v2_3_0_frame_header_parse(struct v2_frame_header *header, const char
 	int i;
 
 	for (i = 0; i < 4; i++) {
-		if (!is_frame_id_char(buf[i]))
+		if (!(is_frame_id_char(buf[i]) || (i == 3 && buf[i] == '\0')))
 			return 0;
 		header->id[i] = buf[i];
 	}
@@ -433,7 +483,7 @@ static int v2_4_0_frame_header_parse(struct v2_frame_header *header, const char
 	int i;
 
 	for (i = 0; i < 4; i++) {
-		if (!is_frame_id_char(buf[i]))
+		if (!(is_frame_id_char(buf[i]) || (i == 3 && buf[i] == '\0')))
 			return 0;
 		header->id[i] = buf[i];
 	}
@@ -447,23 +497,6 @@ static int v2_4_0_frame_header_parse(struct v2_frame_header *header, const char
 	return 1;
 }
 
-static int read_all(int fd, char *buf, size_t size)
-{
-	size_t pos = 0;
-
-	while (pos < size) {
-		int rc = read(fd, buf + pos, size - pos);
-
-		if (rc == -1) {
-			if (errno == EINTR || errno == EAGAIN)
-				continue;
-			return -1;
-		}
-		pos += rc;
-	}
-	return 0;
-}
-
 static char *parse_genre(const char *str)
 {
 	int parenthesis = 0;
@@ -513,8 +546,9 @@ static struct {
 	/* 2.4.0 */
 	{ "TDRC", ID3_DATE }, // recording date
 	{ "TDRL", ID3_DATE }, // release date
-	{ "TDOR", ID3_DATE }, // original release date
+	{ "TDOR", ID3_ORIGINALDATE }, // original release date
 	{ "TSOP", ID3_ARTISTSORT },
+	{ "TSOA", ID3_ALBUMSORT },
 
 	/* >= 2.3.0 */
 	{ "TPE1", ID3_ARTIST },
@@ -525,31 +559,86 @@ static struct {
 	{ "TPOS", ID3_DISC },
 	{ "TRCK", ID3_TRACK },
 	{ "TPE2", ID3_ALBUMARTIST },
+	{ "TSO2", ID3_ALBUMARTISTSORT },
 	{ "XSOP", ID3_ARTISTSORT }, // obsolete
+	{ "XSOA", ID3_ALBUMSORT }, // obsolete
+	{ "TCMP", ID3_COMPILATION },
+	{ "TORY", ID3_ORIGINALDATE },
+	{ "TCOM", ID3_COMPOSER },
+	{ "TPE3", ID3_CONDUCTOR },
+	{ "TEXT", ID3_LYRICIST },
+	{ "TPE4", ID3_REMIXER },
+	{ "TPUB", ID3_PUBLISHER }, // TPUB can be both publisher or label
+	{ "TIT3", ID3_SUBTITLE },
+	{ "TMED", ID3_MEDIA },
 
 	/* obsolete frames (2.2.0) */
 	{ "TP1",  ID3_ARTIST },
+	{ "TP2",  ID3_ALBUMARTIST },
 	{ "TAL",  ID3_ALBUM },
 	{ "TT2",  ID3_TITLE },
 	{ "TYE",  ID3_DATE },
 	{ "TCO",  ID3_GENRE },
 	{ "TPA",  ID3_DISC },
 	{ "TRK",  ID3_TRACK },
-
-	{ "", -1 }
+	{ "TSP",  ID3_ARTISTSORT },
+	{ "TS2",  ID3_ALBUMARTISTSORT },
+	{ "TSA",  ID3_ALBUMSORT },
+	{ "TCP",  ID3_COMPILATION },
 };
 
 static int frame_tab_index(const char *id)
 {
-	int i;
+	int i = 0;
 
-	for (i = 0; frame_tab[i].key != -1; i++) {
+	while (i < N_ELEMENTS(frame_tab)) {
 		if (!strncmp(id, frame_tab[i].name, 4))
 			return i;
+		i++;
 	}
 	return -1;
 }
 
+static int check_date_format(const char *buf)
+{
+	int i, ch;
+
+	/* year */
+	for (i = 0; i < 4; i++) {
+		ch = *buf++;
+		if (ch < '0' || ch > '9')
+			return 0;
+	}
+	ch = *buf++;
+	if (!ch)
+		return 4;
+	if (ch != '-')
+		return 0;
+
+	/* month */
+	for (i = 0; i < 2; i++) {
+		ch = *buf++;
+		if (ch < '0' || ch > '9')
+			return 0;
+	}
+	ch = *buf++;
+	if (!ch)
+		return 7;
+	if (ch != '-')
+		return 0;
+
+	/* day */
+	for (i = 0; i < 2; i++) {
+		ch = *buf++;
+		if (ch < '0' || ch > '9')
+			return 0;
+	}
+	ch = *buf;
+	if (!ch || (ch >= '0' && ch <= '9'))
+		return 10;
+	return 0;
+}
+
 static void fix_date(char *buf)
 {
 	const char *ptr = buf;
@@ -575,115 +664,347 @@ static void fix_date(char *buf)
 static char *decode_str(const char *buf, int len, int encoding)
 {
 	char *in, *out = NULL;
-	int rc = 0;
 
 	switch (encoding) {
-	case 0x00: /* ISO-8859-1 */
+	case ID3_ENCODING_ISO_8859_1:
 		in = xstrndup(buf, len);
-		rc = utf8_encode(in, id3_default_charset, &out);
+		utf8_encode(in, id3_default_charset, &out);
 		free(in);
 		break;
-	case 0x03: /* UTF-8 */
+	case ID3_ENCODING_UTF_8:
 		in = xstrndup(buf, len);
 		if (u_is_valid(in)) {
 			out = in;
 		} else {
-			rc = utf8_encode(in, id3_default_charset, &out);
+			utf8_encode(in, id3_default_charset, &out);
 			free(in);
 		}
 		break;
-	case 0x01: /* UTF-16 */
+	case ID3_ENCODING_UTF_16:
+	case ID3_ENCODING_UTF_16_BE:
 		out = utf16_to_utf8((const unsigned char *)buf, len);
 		break;
-	case 0x02: /* UTF-16BE */
-		out = utf16be_to_utf8((const unsigned char *)buf, len);
-		break;
 	}
 	return out;
 }
 
+static void add_v2(struct id3tag *id3, enum id3_key key, char *value)
+{
+	free(id3->v2[key]);
+	id3->v2[key] = value;
+	id3->has_v2 = 1;
+}
+
+static void decode_normal(struct id3tag *id3, const char *buf, int len, int encoding, enum id3_key key)
+{
+	char *out = decode_str(buf, len, encoding);
+
+	if (!out)
+		return;
+
+	if (key == ID3_GENRE) {
+		char *tmp;
+
+		id3_debug("genre before: '%s'\n", out);
+		tmp = parse_genre(out);
+		free(out);
+		out = tmp;
+	} else if (key == ID3_DATE || key == ID3_ORIGINALDATE) {
+		int date_len = check_date_format(out);
+		id3_debug("date before: '%s'\n", out);
+		if (date_len)
+			out[date_len] = '\0';
+		else
+			fix_date(out);
+		if (!*out) {
+			id3_debug("date parsing failed\n");
+			free(out);
+			return;
+		}
+	} else if (key == ID3_ALBUMARTIST) {
+		/*
+		 * This must be TPE2 frame; ignore it if ID3_ALBUMARTIST is
+		 * already present
+		 */
+		if (id3->v2[key])
+			return;
+	} else if (key == ID3_PUBLISHER) {
+		 add_v2(id3, ID3_LABEL, strdup(out));
+	}
+
+	add_v2(id3, key, out);
+}
+
+static size_t id3_skiplen(const char *buf, size_t len, int encoding)
+{
+	if (encoding == ID3_ENCODING_ISO_8859_1 || encoding == ID3_ENCODING_UTF_8) {
+		return strlen(buf) + 1;
+	} else {
+		int i = 0;
+		while (i + 1 < len) {
+			if (buf[i] == '\0' && buf[i + 1] == '\0')
+				return i + 2;
+
+			/* Assume every character is exactly 2 bytes */
+			i += 2;
+		}
+
+		return len;
+	}
+}
+
+static void decode_txxx(struct id3tag *id3, const char *buf, int len, int encoding)
+{
+	const char ql_prefix[] = "QuodLibet::";
+	enum id3_key key = NUM_ID3_KEYS;
+	int size;
+	char *out, *out_mem;
+
+	out = decode_str(buf, len, encoding);
+	if (!out)
+		return;
+
+	id3_debug("TXXX, key = '%s'\n", out);
+
+	out_mem = out;
+
+	/* skip braindead QuodLibet TXXX frame prefix */
+	if (!strncmp(out, ql_prefix, sizeof(ql_prefix) - 1))
+		out += sizeof(ql_prefix) - 1;
+
+	if (!strcasecmp(out, "replaygain_track_gain"))
+		key = ID3_RG_TRACK_GAIN;
+	else if (!strcasecmp(out, "replaygain_track_peak"))
+		key = ID3_RG_TRACK_PEAK;
+	else if (!strcasecmp(out, "replaygain_album_gain"))
+		key = ID3_RG_ALBUM_GAIN;
+	else if (!strcasecmp(out, "replaygain_album_peak"))
+		key = ID3_RG_ALBUM_PEAK;
+	else if (!strcasecmp(out, "album artist"))
+		key = ID3_ALBUMARTIST;
+	else if (!strcasecmp(out, "albumartist"))
+		key = ID3_ALBUMARTIST;
+	else if (!strcasecmp(out, "albumartistsort"))
+		key = ID3_ALBUMARTISTSORT;
+	else if (!strcasecmp(out, "albumsort"))
+		key = ID3_ALBUMSORT;
+	else if (!strcasecmp(out, "compilation"))
+		key = ID3_COMPILATION;
+
+	size = id3_skiplen(buf, len, encoding);
+	free(out_mem);
+
+	if (key == NUM_ID3_KEYS)
+		return;
+
+	buf += size;
+	len -= size;
+	if (len <= 0)
+		return;
+
+	out = decode_str(buf, len, encoding);
+	if (!out)
+		return;
+
+	add_v2(id3, key, out);
+}
+
+static void decode_comment(struct id3tag *id3, const char *buf, int len, int encoding)
+{
+	int slen;
+	char *out;
+	int valid_description;
+
+	if (len <= 3)
+		return;
+
+	/* skip language */
+	buf += 3;
+	len -= 3;
+
+	/* "Short content description" part of COMM frame */
+	out = decode_str(buf, len, encoding);
+	if (!out)
+		return;
+
+	valid_description = strcmp(out, "") == 0 || strcmp(out, "description") == 0;
+	free(out);
+
+	if (!valid_description)
+		return;
+
+	slen = id3_skiplen(buf, len, encoding);
+	if (slen >= len)
+		return;
+
+	buf += slen;
+	len -= slen;
+
+	out = decode_str(buf, len, encoding);
+	if (!out)
+		return;
+
+	add_v2(id3, ID3_COMMENT, out);
+}
+
+/*
+ * From http://id3.org/id3v2.4.0-frames:
+ *
+ * The volume adjustment is encoded as a fixed point decibel value, 16 bit signed
+ * integer representing (adjustment*512), giving +/- 64 dB with a precision of
+ * 0.001953125 dB. E.g. +2 dB is stored as $04 00 and -2 dB is $FC 00. There may
+ * be more than one "RVA2" frame in each tag, but only one with the same
+ * identification string.
+ *
+ * 	<Header for 'Relative volume adjustment (2)', ID: "RVA2">
+ * 	Identification          <text string> $00
+ *
+ * The 'identification' string is used to identify the situation and/or device
+ * where this adjustment should apply. The following is then repeated for every
+ * channel
+ *
+ * 	Type of channel         $xx
+ * 	Volume adjustment       $xx xx
+ * 	Bits representing peak  $xx
+ * 	Peak volume             $xx (xx ...)
+ *
+ * Type of channel:	$00 Other
+ * 			$01 Master volume
+ * 			$02 Front right
+ * 			$03 Front left
+ * 			$04 Back right
+ * 			$05 Back left
+ * 			$06 Front centre
+ * 			$07 Back centre
+ * 			$08 Subwoofer
+ *
+ * Bits representing peak can be any number between 0 and 255. 0 means that there
+ * is no peak volume field. The peak volume field is always padded to whole
+ * bytes, setting the most significant bits to zero.
+ */
+static void decode_rva2(struct id3tag *id3, const char *buf, int len)
+{
+	const int rva2_min_len	= 6 + 1 + 2 + 1;
+
+	int audiophile_rg	= 0;
+	int channel		= 0;
+	int16_t volume_adj	= 0;
+	int peak_bits		= 0;
+	int peak_bytes		= 0;
+	int peak_shift		= 0;
+	uint32_t peak		= 0;
+
+	char *gain_str		= NULL;
+	char *peak_str		= NULL;
+
+	int i;
+
+	if (len < rva2_min_len) {
+		id3_debug("frame length %d too small\n", len);
+		return;
+	}
+
+	if (!strcasecmp(buf, "album")) {
+		audiophile_rg = 1;
+	} else if (strcasecmp(buf, "track")) {
+		id3_debug("unsupported identifier: %s\n", buf);
+		return;
+	}
+
+	buf += 6;
+
+	channel = *buf++;
+	if (channel != 0x1) {
+		id3_debug("unsupported channel: %d\n", channel);
+		return;
+	}
+
+	get_i16((unsigned char *)buf, &volume_adj);
+	buf += 2;
+
+	peak_bits = *buf++;
+
+	if (peak_bits == 0)
+		id3_debug("no peak data\n");
+
+	/*
+	 * This crazy code comes from Mutagen
+	 */
+	peak_bytes = min(4, (peak_bits + 7) >> 3);
+	peak_shift = ((8 - (peak_bits & 7)) & 7) + (4 - peak_bytes) * 8;
+
+	if (len < rva2_min_len + peak_bytes) {
+		id3_debug("peak data %d does not fit frame with length %d\n", peak_bytes, len);
+		return;
+	}
+
+	for (i = 0; i < peak_bytes; ++i) {
+		peak <<= 8;
+		peak |= (unsigned char)*buf++;
+	}
+
+	gain_str = xnew(char, 32);
+	snprintf(gain_str, 32, "%lf dB", volume_adj / 512.0);
+
+	add_v2(id3, audiophile_rg ? ID3_RG_ALBUM_GAIN : ID3_RG_TRACK_GAIN, gain_str);
+
+	if (peak_bytes) {
+		peak_str = xnew(char, 32);
+		snprintf(peak_str, 32, "%lf", ((double)peak * (1 << peak_shift)) / INT_MAX);
+
+		add_v2(id3, audiophile_rg ? ID3_RG_ALBUM_PEAK : ID3_RG_TRACK_PEAK, peak_str);
+	}
+
+	id3_debug("gain %s, peak %s\n", gain_str, peak_str ? peak_str : "none");
+}
+
+static void decode_ufid(struct id3tag *id3, const char *buf, int len)
+{
+	char *ufid;
+	int ufid_len = len - 22 - 1;
+
+	if (ufid_len < 0 || strcmp(buf, "http://musicbrainz.org") != 0)
+		return;
+
+	ufid = xnew(char, ufid_len + 1);
+	memcpy(ufid, buf + len - ufid_len, ufid_len);
+	ufid[ufid_len] = '\0';
+
+	id3_debug("%s: %s\n", buf, ufid);
+	add_v2(id3, ID3_MUSICBRAINZ_TRACKID, ufid);
+}
+
+
 static void v2_add_frame(struct id3tag *id3, struct v2_frame_header *fh, const char *buf)
 {
-	int idx, encoding = *buf++, len = fh->size - 1;
-	enum id3_key key = NUM_ID3_KEYS;
-	char *out;
+	int encoding;
+	int len;
+	int idx;
 
-	if (encoding > 3)
+	if (!strncmp(fh->id, "RVA2", 4)) {
+		decode_rva2(id3, buf, fh->size);
+		return;
+	} else if (!strncmp(fh->id, "UFID", 4)) {
+		decode_ufid(id3, buf, fh->size);
+		return;
+	}
+
+	encoding = *buf++;
+	len = fh->size - 1;
+
+	if (encoding > ID3_ENCODING_MAX)
 		return;
 
 	idx = frame_tab_index(fh->id);
 	if (idx >= 0) {
-		key = frame_tab[idx].key;
-		out = decode_str(buf, len, encoding);
-		if (!out)
-			return;
-
-		if (key == ID3_GENRE) {
-			char *tmp;
-
-			id3_debug("genre before: '%s'\n", out);
-			tmp = parse_genre(out);
-			free(out);
-			out = tmp;
-		}
-		if (key == ID3_DATE) {
-			id3_debug("date before: '%s'\n", out);
-			fix_date(out);
-			if (!*out) {
-				id3_debug("date parsing failed\n");
-				free(out);
-				return;
-			}
-		}
-
-		id3_debug("%s '%s'\n", frame_tab[idx].name, out);
+		decode_normal(id3, buf, len, encoding, frame_tab[idx].key);
 	} else if (!strncmp(fh->id, "TXXX", 4)) {
-		int size;
-
-		id3_debug("TXXX\n");
-
-		/* TXXX<len><encoding><key><val> */
-		out = decode_str(buf, len, encoding);
-		if (!out)
-			return;
-
-		id3_debug("TXXX, key = '%s'\n", out);
-		if (!strcasecmp(out, "replaygain_track_gain"))
-			key = ID3_RG_TRACK_GAIN;
-		if (!strcasecmp(out, "replaygain_track_peak"))
-			key = ID3_RG_TRACK_PEAK;
-		if (!strcasecmp(out, "replaygain_album_gain"))
-			key = ID3_RG_ALBUM_GAIN;
-		if (!strcasecmp(out, "replaygain_album_peak"))
-			key = ID3_RG_ALBUM_PEAK;
-		if (!strcasecmp(out, "album artist"))
-			key = ID3_ALBUMARTIST;
-		if (!strcasecmp(out, "albumartistsort"))
-			key = ID3_ALBUMARTISTSORT;
-
-		size = strlen(out) + 1;
-		free(out);
-
-		if (key == NUM_ID3_KEYS)
-			return;
-
-		buf += size;
-		len -= size;
-		if (len <= 0)
-			return;
-
-		out = decode_str(buf, len, encoding);
-		if (!out)
-			return;
-
-		id3_debug("TXXX, val = '%s'\n", out);
-	} else {
-		return;
+		decode_txxx(id3, buf, len, encoding);
+	} else if (!strncmp(fh->id, "COMM", 4)) {
+		decode_comment(id3, buf, len, encoding);
+	} else if (!strncmp(fh->id, "COM", 3)) {
+		decode_comment(id3, buf, len, encoding);
 	}
-
-	free(id3->v2[key]);
-	id3->v2[key] = out;
-	id3->has_v2 = 1;
 }
 
 static void unsync(unsigned char *buf, int *lenp)
@@ -724,7 +1045,7 @@ static int v2_read(struct id3tag *id3, int fd, const struct v2_header *header)
 	buf_size = header->size;
 	buf = xnew(char, buf_size);
 	rc = read_all(fd, buf, buf_size);
-	if (rc) {
+	if (rc == -1) {
 		free(buf);
 		return rc;
 	}
@@ -733,8 +1054,7 @@ static int v2_read(struct id3tag *id3, int fd, const struct v2_header *header)
 	if (header->flags & V2_HEADER_EXTENDED) {
 		struct v2_extended_header ext;
 
-		v2_extended_header_parse(&ext, buf);
-		if (ext.size > buf_size) {
+		if (!v2_extended_header_parse(&ext, buf) || ext.size > buf_size) {
 			id3_debug("extended header corrupted\n");
 			free(buf);
 			return -2;
@@ -743,13 +1063,6 @@ static int v2_read(struct id3tag *id3, int fd, const struct v2_header *header)
 		/* should check if update flag is set */
 	}
 
-	if (header->flags & V2_HEADER_UNSYNC) {
-		int len = buf_size - frame_start;
-
-		unsync((unsigned char *)(buf + frame_start), &len);
-		buf_size = len + frame_start;
-	}
-
 	frame_header_size = 10;
 	if (header->ver_major == 2)
 		frame_header_size = 6;
@@ -757,7 +1070,7 @@ static int v2_read(struct id3tag *id3, int fd, const struct v2_header *header)
 	i = frame_start;
 	while (i < buf_size - frame_header_size) {
 		struct v2_frame_header fh;
-		int len;
+		int len_unsync;
 
 		if (header->ver_major == 2) {
 			if (!v2_2_0_frame_header_parse(&fh, buf + i))
@@ -772,20 +1085,28 @@ static int v2_read(struct id3tag *id3, int fd, const struct v2_header *header)
 		}
 
 		i += frame_header_size;
+
 		if (fh.size > buf_size - i) {
 			id3_debug("frame too big\n");
 			break;
 		}
 
-		len = fh.size;
-		if (fh.flags & V2_FRAME_UNSYNC) {
-			int tmp = len;
-
-			unsync((unsigned char *)(buf + i), &tmp);
-			fh.size = tmp;
+		if (fh.flags & V2_FRAME_LEN_INDICATOR) {
+			/*
+			 * Ignore the frame length 4-byte field
+			 */
+			i	+= 4;
+			fh.size	-= 4;
 		}
+
+		len_unsync = fh.size;
+
+		if ((fh.flags & V2_FRAME_UNSYNC) || (header->flags & V2_HEADER_UNSYNC))
+			unsync((unsigned char *)(buf + i), (int *)&fh.size);
+
 		v2_add_frame(id3, &fh, buf + i);
-		i += len;
+
+		i += len_unsync;
 	}
 
 	free(buf);
@@ -815,6 +1136,12 @@ int id3_tag_size(const char *buf, int buf_size)
 	return 0;
 }
 
+void id3_init(struct id3tag *id3)
+{
+	const struct id3tag t = { .has_v1 = 0, .has_v2 = 0 };
+	*id3 = t;
+}
+
 void id3_free(struct id3tag *id3)
 {
 	int i;
@@ -833,7 +1160,7 @@ int id3_read_tags(struct id3tag *id3, int fd, unsigned int flags)
 		char buf[138];
 
 		rc = read_all(fd, buf, 10);
-		if (rc)
+		if (rc == -1)
 			goto rc_error;
 		if (v2_header_parse(&header, buf)) {
 			rc = v2_read(id3, fd, &header);
@@ -847,7 +1174,7 @@ int id3_read_tags(struct id3tag *id3, int fd, unsigned int flags)
 			if (off == -1)
 				goto error;
 			rc = read_all(fd, buf, 138);
-			if (rc)
+			if (rc == -1)
 				goto rc_error;
 
 			if (is_v1(buf + 10)) {
@@ -857,7 +1184,7 @@ int id3_read_tags(struct id3tag *id3, int fd, unsigned int flags)
 				}
 				if (v2_footer_parse(&header, buf)) {
 					/* footer at end of file - 128 */
-					off = lseek(fd, -(header.size + 138), SEEK_END);
+					off = lseek(fd, -((off_t) header.size + 138), SEEK_END);
 					if (off == -1)
 						goto error;
 					rc = v2_read(id3, fd, &header);
@@ -866,7 +1193,7 @@ int id3_read_tags(struct id3tag *id3, int fd, unsigned int flags)
 				}
 			} else if (v2_footer_parse(&header, buf + 128)) {
 				/* footer at end of file */
-				off = lseek(fd, -(header.size + 10), SEEK_END);
+				off = lseek(fd, -((off_t) header.size + 10), SEEK_END);
 				if (off == -1)
 					goto error;
 				rc = v2_read(id3, fd, &header);
@@ -881,7 +1208,7 @@ int id3_read_tags(struct id3tag *id3, int fd, unsigned int flags)
 		if (off == -1)
 			goto error;
 		rc = read_all(fd, id3->v1, 128);
-		if (rc)
+		if (rc == -1)
 			goto rc_error;
 		id3->has_v1 = is_v1(id3->v1);
 	}
@@ -953,3 +1280,10 @@ char *id3_get_comment(struct id3tag *id3, enum id3_key key)
 	}
 	return NULL;
 }
+
+char const *id3_get_genre(uint16_t id)
+{
+	if (id >= NR_GENRES)
+		return NULL;
+	return genres[id];
+}
