@@ -1,5 +1,5 @@
 /*
- * Copyright 2008-2011 Various Authors
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -16,6 +16,7 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#include "convert.h"
 #include "input.h"
 #include "ip.h"
 #include "pcm.h"
@@ -70,8 +71,8 @@ struct input_plugin {
 	 * pcm is converted to 16-bit signed little-endian stereo
 	 * NOTE: no conversion is done if channels > 2 or bits > 16
 	 */
-	void (*pcm_convert)(char *, const char *, int);
-	void (*pcm_convert_in_place)(char *, int);
+	void (*pcm_convert)(void *, const void *, int);
+	void (*pcm_convert_in_place)(void *, int);
 	/*
 	 * 4  if 8-bit mono
 	 * 2  if 8-bit stereo or 16-bit mono
@@ -272,15 +273,15 @@ static int setup_remote(struct input_plugin *ip, const struct keyval *headers, i
 
 	val = keyvals_get_val(headers, "icy-name");
 	if (val)
-		ip->data.icy_name = xstrdup(val);
+		ip->data.icy_name = to_utf8(val, icecast_default_charset);
 
 	val = keyvals_get_val(headers, "icy-genre");
 	if (val)
-		ip->data.icy_genre = xstrdup(val);
+		ip->data.icy_genre = to_utf8(val, icecast_default_charset);
 
 	val = keyvals_get_val(headers, "icy-url");
 	if (val)
-		ip->data.icy_url = xstrdup(val);
+		ip->data.icy_url = to_utf8(val, icecast_default_charset);
 
 	return 0;
 }
