@@ -1,5 +1,5 @@
 /*
- * Copyright 2008-2011 Various Authors
+ * Copyright 2008-2013 Various Authors
  * Copyright 2008 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -16,8 +16,8 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
-#ifndef JOB_H
-#define JOB_H
+#ifndef CMUS_JOB_H
+#define CMUS_JOB_H
 
 #include "cmus.h"
 
@@ -25,6 +25,7 @@ struct add_data {
 	enum file_type type;
 	char *name;
 	add_ti_cb add;
+	unsigned int force : 1;
 };
 
 struct update_data {
