@@ -1,11 +1,27 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2005-2006 Timo Hirvonen
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#include "misc.h"
 #include "prog.h"
 #include "file.h"
 #include "path.h"
 #include "xmalloc.h"
+#include "utils.h"
 
 #include <unistd.h>
 #include <fcntl.h>
@@ -24,32 +40,49 @@ static int sock;
 static int raw_args = 0;
 static char *passwd;
 
-static void gethostbyname_failed(void)
+static int read_answer(void)
 {
-	const char *error = "Unknown error.";
+	char buf[8192];
+	int got_nl = 0;
+	int len = 0;
 
-	switch (h_errno) {
-	case HOST_NOT_FOUND:
-	case NO_ADDRESS:
-		error = "Host not found.";
-		break;
-	case NO_RECOVERY:
-		error = "A non-recoverable name server error.";
-		break;
-	case TRY_AGAIN:
-		error = "A temporary error occurred on an authoritative name server.";
-		break;
+	while (1) {
+		int rc = read(sock, buf, sizeof(buf));
+
+		if (rc < 0) {
+			warn_errno("read");
+			break;
+		}
+		if (!rc)
+			die("unexpected EOF\n");
+
+		len += rc;
+
+		// Last line should be empty (i.e. read "\n" or "...\n\n").
+		// Write everything but the last \n to stdout.
+		if (got_nl && buf[0] == '\n')
+			break;
+		if (len == 1 && buf[0] == '\n')
+			break;
+		if (rc > 1 && buf[rc - 1] == '\n' && buf[rc - 2] == '\n') {
+			write_all(1, buf, rc - 1);
+			break;
+		}
+		got_nl = buf[rc - 1] == '\n';
+		write_all(1, buf, rc);
 	}
-	die("gethostbyname: %s\n", error);
+	return len;
 }
 
-static void write_line(const char *line)
+static int write_line(const char *line)
 {
 	if (write_all(sock, line, strlen(line)) == -1)
 		die_errno("write");
+
+	return read_answer();
 }
 
-static void send_cmd(const char *format, ...)
+static int send_cmd(const char *format, ...)
 {
 	char buf[512];
 	va_list ap;
@@ -58,17 +91,17 @@ static void send_cmd(const char *format, ...)
 	vsnprintf(buf, sizeof(buf), format, ap);
 	va_end(ap);
 
-	write_line(buf);
+	return write_line(buf);
 }
 
-static void remote_connect(const char *address)
+static int remote_connect(const char *address)
 {
 	union {
 		struct sockaddr sa;
 		struct sockaddr_un un;
-		struct sockaddr_in in;
+		struct sockaddr_storage sas;
 	} addr;
-	int addrlen;
+	size_t addrlen;
 
 	if (strchr(address, '/')) {
 		if (passwd)
@@ -80,32 +113,28 @@ static void remote_connect(const char *address)
 
 		addrlen = sizeof(addr.un);
 	} else {
-		char *s = strchr(address, ':');
-		int port = DEFAULT_PORT;
-		struct hostent *hent;
+		const struct addrinfo hints = {
+			.ai_socktype = SOCK_STREAM
+		};
+		const char *port = STRINGIZE(DEFAULT_PORT);
+		struct addrinfo *result;
+		char *s = strrchr(address, ':');
+		int rc;
 
 		if (!passwd)
 			die("password required for tcp/ip connection\n");
 
 		if (s) {
 			*s++ = 0;
-			port = atoi(s);
+			port = s;
 		}
-		hent = gethostbyname(address);
-		if (!hent)
-			gethostbyname_failed();
 
-		addr.sa.sa_family = hent->h_addrtype;
-		switch (addr.sa.sa_family) {
-		case AF_INET:
-			memcpy(&addr.in.sin_addr, hent->h_addr_list[0], hent->h_length);
-			addr.in.sin_port = htons(port);
-
-			addrlen = sizeof(addr.in);
-			break;
-		default:
-			die("unsupported address type\n");
-		}
+		rc = getaddrinfo(address, port, &hints, &result);
+		if (rc != 0)
+			die("getaddrinfo: %s\n", gai_strerror(rc));
+		memcpy(&addr.sa, result->ai_addr, result->ai_addrlen);
+		addrlen = result->ai_addrlen;
+		freeaddrinfo(result);
 	}
 
 	sock = socket(addr.sa.sa_family, SOCK_STREAM, 0);
@@ -123,7 +152,8 @@ static void remote_connect(const char *address)
 	}
 
 	if (passwd)
-		send_cmd("%s\n", passwd);
+		return send_cmd("passwd %s\n", passwd) == 1;
+	return 1;
 }
 
 static char *file_url_absolute(const char *str)
@@ -155,6 +185,7 @@ enum flags {
 	FLAG_SHUFFLE,
 	FLAG_VOLUME,
 	FLAG_SEEK,
+	FLAG_QUERY,
 
 	FLAG_LIBRARY,
 	FLAG_PLAYLIST,
@@ -181,6 +212,7 @@ static struct option options[NR_FLAGS + 1] = {
 	{ 'S', "shuffle", 0 },
 	{ 'v', "volume", 1 },
 	{ 'k', "seek", 1 },
+	{ 'Q', "query", 0 },
 
 	{ 'l', "library", 0 },
 	{ 'P', "playlist", 0 },
@@ -199,7 +231,7 @@ static const char *usage =
 "   or: %s\n"
 "Control cmus through socket.\n"
 "\n"
-"      --server ADDR    connect using ADDR instead of ~/.cmus/socket\n"
+"      --server ADDR    connect using ADDR instead of $CMUS_SOCKET or $XDG_RUNTIME_DIR/cmus-socket\n"
 "                       ADDR is either a UNIX socket or host[:port]\n"
 "                       WARNING: using TCP/IP is insecure!\n"
 "      --passwd PASSWD  password to use for TCP/IP connection\n"
@@ -217,6 +249,7 @@ static const char *usage =
 "  -S, --shuffle        toggle shuffle\n"
 "  -v, --volume VOL     vol VOL\n"
 "  -k, --seek SEEK      seek SEEK\n"
+"  -Q, --query          get player status (same as -C status)\n"
 "\n"
 "  -l, --library        modify library instead of playlist\n"
 "  -P, --playlist       modify playlist (default)\n"
@@ -234,11 +267,11 @@ static const char *usage =
 
 int main(int argc, char *argv[])
 {
-	char server_buf[256];
 	char *server = NULL;
 	char *play_file = NULL;
 	char *volume = NULL;
 	char *seek = NULL;
+	int query = 0;
 	int i, nr_cmds = 0;
 	int context = 'p';
 
@@ -258,7 +291,9 @@ int main(int argc, char *argv[])
 			printf(usage, program_name, program_name, program_name);
 			return 0;
 		case FLAG_VERSION:
-			printf("cmus " VERSION "\nCopyright 2004-2006 Timo Hirvonen\n");
+			printf("cmus " VERSION
+			       "\nCopyright 2004-2006 Timo Hirvonen"
+			       "\nCopyright 2008-2015 Various Authors\n");
 			return 0;
 		case FLAG_SERVER:
 			server = arg;
@@ -274,6 +309,10 @@ int main(int argc, char *argv[])
 			seek = arg;
 			nr_cmds++;
 			break;
+		case FLAG_QUERY:
+			query = 1;
+			nr_cmds++;
+			break;
 		case FLAG_FILE:
 			play_file = arg;
 			nr_cmds++;
@@ -306,22 +345,12 @@ int main(int argc, char *argv[])
 	if (nr_cmds && raw_args)
 		die("don't mix raw and cooked stuff\n");
 
-	if (server == NULL) {
-		const char *config_dir = getenv("CMUS_HOME");
+	misc_init();
+	if (server == NULL)
+		server = xstrdup(cmus_socket_path);
 
-		if (config_dir && config_dir[0]) {
-			snprintf(server_buf, sizeof(server_buf), "%s/socket", config_dir);
-		} else {
-			const char *home = getenv("HOME");
-
-			if (!home)
-				die("error: environment variable HOME not set\n");
-			snprintf(server_buf, sizeof(server_buf), "%s/.cmus/socket", home);
-		}
-		server = server_buf;
-	}
-
-	remote_connect(server);
+	if (!remote_connect(server))
+		return 1;
 
 	if (raw_args) {
 		while (*argv)
@@ -365,5 +394,7 @@ int main(int argc, char *argv[])
 		send_cmd("vol %s\n", volume);
 	if (seek)
 		send_cmd("seek %s\n", seek);
+	if (query)
+		send_cmd("status\n");
 	return 0;
 }
