@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,11 +11,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "mixer.h"
@@ -22,6 +21,8 @@
 #include "xmalloc.h"
 #include "debug.h"
 
+#include <strings.h>
+
 #define ALSA_PCM_NEW_HW_PARAMS_API
 #define ALSA_PCM_NEW_SW_PARAMS_API
 
@@ -54,15 +55,43 @@ static int alsa_mixer_exit(void)
 	return 0;
 }
 
+static snd_mixer_elem_t *find_mixer_elem_by_name(const char *goal_name)
+{
+	snd_mixer_elem_t *elem;
+	snd_mixer_selem_id_t *sid = NULL;
+
+	snd_mixer_selem_id_malloc(&sid);
+
+	for (elem = snd_mixer_first_elem(alsa_mixer_handle); elem;
+		 elem = snd_mixer_elem_next(elem)) {
+
+		const char *name;
+
+		snd_mixer_selem_get_id(elem, sid);
+		name = snd_mixer_selem_id_get_name(sid);
+		d_print("name = %s\n", name);
+		d_print("has playback volume = %d\n", snd_mixer_selem_has_playback_volume(elem));
+		d_print("has playback switch = %d\n", snd_mixer_selem_has_playback_switch(elem));
+
+		if (strcasecmp(name, goal_name) == 0) {
+			if (!snd_mixer_selem_has_playback_volume(elem)) {
+				d_print("mixer element `%s' does not have playback volume\n", name);
+				elem = NULL;
+			}
+			break;
+		}
+	}
+
+	snd_mixer_selem_id_free(sid);
+	return elem;
+}
+
 static int alsa_mixer_open(int *volume_max)
 {
-	snd_mixer_selem_id_t *sid;
 	snd_mixer_elem_t *elem;
 	int count;
 	int rc;
 
-	snd_mixer_selem_id_alloca(&sid);
-
 	rc = snd_mixer_open(&alsa_mixer_handle, 0);
 	if (rc < 0)
 		goto error;
@@ -80,35 +109,23 @@ static int alsa_mixer_open(int *volume_max)
 		d_print("error: mixer does not have elements\n");
 		return -2;
 	}
-	elem = snd_mixer_first_elem(alsa_mixer_handle);
-	while (elem) {
-		const char *name;
-		int has_vol, has_switch;
 
-		snd_mixer_selem_get_id(elem, sid);
-		name = snd_mixer_selem_id_get_name(sid);
-		d_print("name = %s\n", name);
-		d_print("has playback volume = %d\n", snd_mixer_selem_has_playback_volume(elem));
-		d_print("has playback switch = %d\n", snd_mixer_selem_has_playback_switch(elem));
-		if (strcasecmp(name, alsa_mixer_element)) {
-			elem = snd_mixer_elem_next(elem);
-			continue;
-		}
-		has_vol = snd_mixer_selem_has_playback_volume(elem);
-		if (!has_vol) {
-			d_print("mixer element `%s' does not have playback volume\n", name);
+	elem = find_mixer_elem_by_name(alsa_mixer_element);
+	if (!elem) {
+		d_print("mixer element `%s' not found, trying `Master'\n", alsa_mixer_element);
+		elem = find_mixer_elem_by_name("Master");
+		if (!elem) {
+			d_print("error: cannot find suitable mixer element\n");
 			return -2;
 		}
-		snd_mixer_selem_get_playback_volume_range(elem,
-				&mixer_vol_min, &mixer_vol_max);
-		has_switch = snd_mixer_selem_has_playback_switch(elem);
-		/* FIXME: get number of channels */
-		mixer_elem = elem;
-		*volume_max = mixer_vol_max - mixer_vol_min;
-		return 0;
 	}
-	d_print("error: mixer element `%s' not found\n", alsa_mixer_element);
-	return -2;
+	snd_mixer_selem_get_playback_volume_range(elem, &mixer_vol_min, &mixer_vol_max);
+	/* FIXME: get number of channels */
+	mixer_elem = elem;
+	*volume_max = mixer_vol_max - mixer_vol_min;
+
+	return 0;
+
 error:
 	d_print("error: %s\n", snd_strerror(rc));
 	return -1;
@@ -120,6 +137,17 @@ static int alsa_mixer_close(void)
 	return 0;
 }
 
+static int alsa_mixer_get_fds(int *fds)
+{
+	struct pollfd pfd[NR_MIXER_FDS];
+	int count, i;
+
+	count = snd_mixer_poll_descriptors(alsa_mixer_handle, pfd, NR_MIXER_FDS);
+	for (i = 0; i < count; i++)
+		fds[i] = pfd[i].fd;
+	return count;
+}
+
 static int alsa_mixer_set_volume(int l, int r)
 {
 	if (mixer_elem == NULL) {
@@ -191,6 +219,7 @@ const struct mixer_plugin_ops op_mixer_ops = {
 	.exit = alsa_mixer_exit,
 	.open = alsa_mixer_open,
 	.close = alsa_mixer_close,
+	.get_fds = alsa_mixer_get_fds,
 	.set_volume = alsa_mixer_set_volume,
 	.get_volume = alsa_mixer_get_volume,
 	.set_option = alsa_mixer_set_option,
