@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,11 +11,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "op.h"
@@ -49,17 +48,34 @@ static int oss_reset(void)
 	return 0;
 }
 
+/* defined only in OSSv4, but seem to work in OSSv3 (Linux) */
+#ifndef AFMT_S32_LE
+#define AFMT_S32_LE	0x00001000
+#endif
+#ifndef AFMT_S32_BE
+#define AFMT_S32_BE	0x00002000
+#endif
+#ifndef AFMT_S24_PACKED
+#define AFMT_S24_PACKED	0x00040000
+#endif
+
 static int oss_set_sf(sample_format_t sf)
 {
 	int tmp, log2_fragment_size, nr_fragments, bytes_per_second;
 
 	oss_reset();
 	oss_sf = sf;
-	tmp = sf_get_channels(oss_sf) - 1;
-	if (ioctl(oss_fd, SNDCTL_DSP_STEREO, &tmp) == -1) {
-		d_print("SNDCTL_DSP_STEREO failed: %s\n", strerror(errno));
+
+#ifdef SNDCTL_DSP_CHANNELS
+	tmp = sf_get_channels(oss_sf);
+	if (ioctl(oss_fd, SNDCTL_DSP_CHANNELS, &tmp) == -1)
 		return -1;
-	}
+#else
+	tmp = sf_get_channels(oss_sf) - 1;
+	if (ioctl(oss_fd, SNDCTL_DSP_STEREO, &tmp) == -1)
+		return -1;
+#endif
+
 	if (sf_get_bits(oss_sf) == 16) {
 		if (sf_get_signed(oss_sf)) {
 			if (sf_get_bigendian(oss_sf)) {
@@ -80,36 +96,38 @@ static int oss_set_sf(sample_format_t sf)
 		} else {
 			tmp = AFMT_U8;
 		}
+	} else if (sf_get_bits(oss_sf) == 32 && sf_get_signed(oss_sf)) {
+		if (sf_get_bigendian(oss_sf)) {
+			tmp = AFMT_S32_BE;
+		} else {
+			tmp = AFMT_S32_LE;
+		}
+	} else if (sf_get_bits(oss_sf) == 24 && sf_get_signed(oss_sf) && !sf_get_bigendian(oss_sf)) {
+		tmp = AFMT_S24_PACKED;
 	} else {
-		d_print("bits is neither 8 nor 16\n");
+		d_print("unsupported sample format: %c%u_%s\n",
+			sf_get_signed(oss_sf) ? 'S' : 'U', sf_get_bits(oss_sf),
+			sf_get_bigendian(oss_sf) ? "BE" : "LE");
 		return -1;
 	}
-	if (ioctl(oss_fd, SNDCTL_DSP_SAMPLESIZE, &tmp) == -1) {
-		d_print("SNDCTL_DSP_SAMPLESIZE failed: %s\n", strerror(errno));
+	if (ioctl(oss_fd, SNDCTL_DSP_SAMPLESIZE, &tmp) == -1)
 		return -1;
-	}
+
 	tmp = sf_get_rate(oss_sf);
-	if (ioctl(oss_fd, SNDCTL_DSP_SPEED, &tmp) == -1) {
-		d_print("SNDCTL_DSP_SPEED failed: %s\n", strerror(errno));
+	if (ioctl(oss_fd, SNDCTL_DSP_SPEED, &tmp) == -1)
 		return -1;
-	}
-#if 1
+
 	bytes_per_second = sf_get_second_size(oss_sf);
 	log2_fragment_size = 0;
 	while (1 << log2_fragment_size < bytes_per_second / 25)
 		log2_fragment_size++;
 	log2_fragment_size--;
 	nr_fragments = 32;
-/* 	device_buffer_size = (1 << log2_fragment_size) * (nr_fragments + 1); */
+
 	/* bits 0..15 = size of fragment, 16..31 = log2(number of fragments) */
 	tmp = (nr_fragments << 16) + log2_fragment_size;
-	d_print("fragment: %d\n", tmp);
-	if (ioctl(oss_fd, SNDCTL_DSP_SETFRAGMENT, &tmp) == -1) {
-		d_print("fragment: %d\n", tmp);
-		d_print("SNDCTL_DSP_SETFRAGMENT failed: %s\n", strerror(errno));
+	if (ioctl(oss_fd, SNDCTL_DSP_SETFRAGMENT, &tmp) == -1)
 		return -1;
-	}
-#endif
 	return 0;
 }
 
@@ -117,12 +135,9 @@ static int oss_device_exists(const char *device)
 {
 	struct stat s;
 
-	if (stat(device, &s) == 0) {
-		d_print("device %s exists\n", device);
-		return 1;
-	}
-	d_print("device %s does not exist\n", device);
-	return 0;
+	if (stat(device, &s))
+		return 0;
+	return 1;
 }
 
 static int oss_init(void)
@@ -155,11 +170,14 @@ static int oss_exit(void)
 	return 0;
 }
 
-static int oss_open(sample_format_t sf)
+static int oss_open(sample_format_t sf, const channel_position_t *channel_map)
 {
+	int oss_version = 0;
 	oss_fd = open(oss_dsp_device, O_WRONLY);
 	if (oss_fd == -1)
 		return -1;
+	ioctl(oss_fd, OSS_GETVERSION, &oss_version);
+	d_print("oss version: %#08x\n", oss_version);
 	if (oss_set_sf(sf) == -1) {
 		oss_close();
 		return -1;
@@ -178,8 +196,7 @@ static int oss_write(const char *buffer, int count)
 {
 	int rc;
 
-	count /= 4;
-	count *= 4;
+	count -= count % sf_get_frame_size(oss_sf);
 	rc = write(oss_fd, buffer, count);
 	return rc;
 }
@@ -203,8 +220,7 @@ static int oss_buffer_space(void)
 
 	if (ioctl(oss_fd, SNDCTL_DSP_GETOSPACE, &info) == -1)
 		return -1;
-	space = info.bytes;
-/* 	d_print("%d\n", space); */
+	space = (info.fragments - 1) * info.fragsize;
 	return space;
 }
 
