@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2006 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,11 +11,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "player.h"
@@ -26,7 +25,9 @@
 #include "xmalloc.h"
 #include "debug.h"
 #include "compiler.h"
+#include "options.h"
 
+#include <stdio.h>
 #include <stdlib.h>
 #include <pthread.h>
 #include <errno.h>
@@ -35,6 +36,8 @@
 #include <stdarg.h>
 #include <math.h>
 
+const char * const player_status_names[] = { "stopped", "playing", "paused", NULL };
+
 enum producer_status {
 	PS_UNLOADED,
 	PS_STOPPED,
@@ -54,8 +57,7 @@ struct player_info player_info = {
 	.metadata = { 0, },
 	.status = PLAYER_STATUS_STOPPED,
 	.pos = 0,
-	.vol_left = 0,
-	.vol_right = 0,
+	.current_bitrate = -1,
 	.buffer_fill = 0,
 	.buffer_size = 0,
 	.error_msg = NULL,
@@ -64,36 +66,45 @@ struct player_info player_info = {
 	.status_changed = 0,
 	.position_changed = 0,
 	.buffer_fill_changed = 0,
-	.vol_changed = 0,
 };
 
 /* continue playing after track is finished? */
 int player_cont = 1;
 
+/* repeat current track forever? */
+int player_repeat_current;
+
 enum replaygain replaygain;
 int replaygain_limit = 1;
-double replaygain_preamp = 6.0;
+double replaygain_preamp = 0.0;
+
+int soft_vol;
+int soft_vol_l;
+int soft_vol_r;
 
 static const struct player_callbacks *player_cbs = NULL;
 
 static sample_format_t buffer_sf;
+static CHANNEL_MAP(buffer_channel_map);
 
 static pthread_t producer_thread;
 static pthread_mutex_t producer_mutex = CMUS_MUTEX_INITIALIZER;
+static pthread_cond_t producer_playing = CMUS_COND_INITIALIZER;
 static int producer_running = 1;
 static enum producer_status producer_status = PS_UNLOADED;
 static struct input_plugin *ip = NULL;
 
 static pthread_t consumer_thread;
 static pthread_mutex_t consumer_mutex = CMUS_MUTEX_INITIALIZER;
+static pthread_cond_t consumer_playing = CMUS_COND_INITIALIZER;
 static int consumer_running = 1;
 static enum consumer_status consumer_status = CS_STOPPED;
-static unsigned int consumer_pos = 0;
+static unsigned long consumer_pos = 0;
 
 /* for replay gain and soft vol
- * usually same as consumer_pos, sometimes less than consumer_pos
+ * usually same as consumer_pos, sometimes more than consumer_pos
  */
-static unsigned int scale_pos;
+static unsigned long scale_pos;
 static double replaygain_scale = 1.0;
 
 /* locking {{{ */
@@ -123,16 +134,22 @@ static void reset_buffer(void)
 	buffer_reset();
 	consumer_pos = 0;
 	scale_pos = 0;
+	pthread_cond_broadcast(&producer_playing);
 }
 
-static void set_buffer_sf(sample_format_t sf)
+static void set_buffer_sf(void)
 {
-	buffer_sf = sf;
+	buffer_sf = ip_get_sf(ip);
+	ip_get_channel_map(ip, buffer_channel_map);
 
 	/* ip_read converts samples to this format */
 	if (sf_get_channels(buffer_sf) <= 2 && sf_get_bits(buffer_sf) <= 16) {
 		buffer_sf &= SF_RATE_MASK;
 		buffer_sf |= sf_channels(2) | sf_bits(16) | sf_signed(1);
+#ifdef WORDS_BIGENDIAN
+		buffer_sf |= sf_bigendian(1);
+#endif
+		channel_map_init_stereo(buffer_channel_map);
 	}
 }
 
@@ -157,27 +174,124 @@ static const unsigned short soft_vol_db[100] = {
 	0xcdf1, 0xd71a, 0xe59c, 0xefd3
 };
 
-static inline void scale_sample(signed short *buf, int i, int vol)
+static inline void scale_sample_int16_t(int16_t *buf, int i, int vol, int swap)
 {
-	int sample = buf[i];
+	int32_t sample = swap ? (int16_t)swap_uint16(buf[i]) : buf[i];
 
 	if (sample < 0) {
 		sample = (sample * vol - SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
-		if (sample < -32768)
-			sample = -32768;
+		if (sample < INT16_MIN)
+			sample = INT16_MIN;
 	} else {
 		sample = (sample * vol + SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
-		if (sample > 32767)
-			sample = 32767;
+		if (sample > INT16_MAX)
+			sample = INT16_MAX;
+	}
+	buf[i] = swap ? swap_uint16(sample) : sample;
+}
+
+static inline int32_t scale_sample_s24le(int32_t s, int vol)
+{
+	int64_t sample = s;
+	if (sample < 0) {
+		sample = (sample * vol - SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
+		if (sample < -0x800000)
+			sample = -0x800000;
+	} else {
+		sample = (sample * vol + SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
+		if (sample > 0x7fffff)
+			sample = 0x7fffff;
+	}
+	return sample;
+}
+
+static inline void scale_sample_int32_t(int32_t *buf, int i, int vol, int swap)
+{
+	int64_t sample = swap ? (int32_t)swap_uint32(buf[i]) : buf[i];
+
+	if (sample < 0) {
+		sample = (sample * vol - SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
+		if (sample < INT32_MIN)
+			sample = INT32_MIN;
+	} else {
+		sample = (sample * vol + SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
+		if (sample > INT32_MAX)
+			sample = INT32_MAX;
+	}
+	buf[i] = swap ? swap_uint32(sample) : sample;
+}
+
+static inline int sf_need_swap(sample_format_t sf)
+{
+#ifdef WORDS_BIGENDIAN
+	return !sf_get_bigendian(sf);
+#else
+	return sf_get_bigendian(sf);
+#endif
+}
+
+#define SCALE_SAMPLES(TYPE, buffer, count, l, r, swap)				\
+{										\
+	const int frames = count / sizeof(TYPE) / 2;				\
+	TYPE *buf = (void *) buffer;						\
+	int i;									\
+	/* avoid underflowing -32768 to 32767 when scale is 65536 */		\
+	if (l != SOFT_VOL_SCALE && r != SOFT_VOL_SCALE) {			\
+		for (i = 0; i < frames; i++) {					\
+			scale_sample_##TYPE(buf, i * 2, l, swap);		\
+			scale_sample_##TYPE(buf, i * 2 + 1, r, swap);		\
+		}								\
+	} else if (l != SOFT_VOL_SCALE) {					\
+		for (i = 0; i < frames; i++)					\
+			scale_sample_##TYPE(buf, i * 2, l, swap);		\
+	} else if (r != SOFT_VOL_SCALE) {					\
+		for (i = 0; i < frames; i++)					\
+			scale_sample_##TYPE(buf, i * 2 + 1, r, swap);		\
+	}									\
+}
+
+static inline int32_t read_s24le(const char *buf)
+{
+	const unsigned char *b = (const unsigned char *) buf;
+	return b[0] | (b[1] << 8) | (((const signed char *) buf)[2] << 16);
+}
+
+static inline void write_s24le(char *buf, int32_t x)
+{
+	unsigned char *b = (unsigned char *) buf;
+	b[0] = x;
+	b[1] = x >> 8;
+	b[2] = x >> 16;
+}
+
+static void scale_samples_s24le(char *buf, unsigned int count, int l, int r)
+{
+	int frames = count / 3 / 2;
+	if (l != SOFT_VOL_SCALE && r != SOFT_VOL_SCALE) {
+		while (frames--) {
+			write_s24le(buf, scale_sample_s24le(read_s24le(buf), l));
+			buf += 3;
+			write_s24le(buf, scale_sample_s24le(read_s24le(buf), r));
+			buf += 3;
+		}
+	} else if (l != SOFT_VOL_SCALE) {
+		while (frames--) {
+			write_s24le(buf, scale_sample_s24le(read_s24le(buf), l));
+			buf += 3 * 2;
+		}
+	} else if (r != SOFT_VOL_SCALE) {
+		buf += 3;
+		while (frames--) {
+			write_s24le(buf, scale_sample_s24le(read_s24le(buf), r));
+			buf += 3 * 2;
+		}
 	}
-	buf[i] = sample;
 }
 
 static void scale_samples(char *buffer, unsigned int *countp)
 {
-	signed short *buf;
 	unsigned int count = *countp;
-	int ch, bits, l, r, i;
+	int ch, bits, l, r;
 
 	BUG_ON(scale_pos < consumer_pos);
 
@@ -190,64 +304,72 @@ static void scale_samples(char *buffer, unsigned int *countp)
 		count -= offs;
 	}
 	scale_pos += count;
-	buf = (signed short *)buffer;
 
 	if (replaygain_scale == 1.0 && soft_vol_l == 100 && soft_vol_r == 100)
 		return;
 
 	ch = sf_get_channels(buffer_sf);
 	bits = sf_get_bits(buffer_sf);
-	if (ch != 2 || bits != 16)
+	if (ch != 2 || (bits != 16 && bits != 24 && bits != 32))
 		return;
 
 	l = SOFT_VOL_SCALE;
 	r = SOFT_VOL_SCALE;
-	if (soft_vol_l != 100)
+	if (soft_vol && soft_vol_l != 100)
 		l = soft_vol_db[soft_vol_l];
-	if (soft_vol_r != 100)
+	if (soft_vol && soft_vol_r != 100)
 		r = soft_vol_db[soft_vol_r];
 
 	l *= replaygain_scale;
 	r *= replaygain_scale;
 
-	for (i = 0; i < count / 4; i++) {
-		scale_sample(buf, i * 2, l);
-		scale_sample(buf, i * 2 + 1, r);
+	switch (bits) {
+	case 16:
+		SCALE_SAMPLES(int16_t, buffer, count, l, r, sf_need_swap(buffer_sf));
+		break;
+	case 24:
+		if (likely(!sf_get_bigendian(buffer_sf)))
+			scale_samples_s24le(buffer, count, l, r);
+		break;
+	case 32:
+		SCALE_SAMPLES(int32_t, buffer, count, l, r, sf_need_swap(buffer_sf));
+		break;
 	}
 }
 
-static int parse_double(const char *str, double *val)
-{
-	char *end;
-
-	*val = strtod(str, &end);
-	return str == end;
-}
-
 static void update_rg_scale(void)
 {
-	const char *g, *p;
 	double gain, peak, db, scale, limit;
 
 	replaygain_scale = 1.0;
 	if (!player_info.ti || !replaygain)
 		return;
 
-	if (replaygain == RG_TRACK) {
-		g = comments_get_val(player_info.ti->comments, "replaygain_track_gain");
-		p = comments_get_val(player_info.ti->comments, "replaygain_track_peak");
+	if (replaygain == RG_TRACK || replaygain == RG_TRACK_PREFERRED) {
+		gain = player_info.ti->rg_track_gain;
+		peak = player_info.ti->rg_track_peak;
 	} else {
-		g = comments_get_val(player_info.ti->comments, "replaygain_album_gain");
-		p = comments_get_val(player_info.ti->comments, "replaygain_album_peak");
+		gain = player_info.ti->rg_album_gain;
+		peak = player_info.ti->rg_album_peak;
 	}
 
-	if (!g || !p) {
-		d_print("gain or peak not available\n");
+	if (isnan(gain)) {
+		if (replaygain == RG_TRACK_PREFERRED) {
+			gain = player_info.ti->rg_album_gain;
+			peak = player_info.ti->rg_album_peak;
+		} else if (replaygain == RG_ALBUM_PREFERRED) {
+			gain = player_info.ti->rg_track_gain;
+			peak = player_info.ti->rg_track_peak;
+		}
+	}
+
+	if (isnan(gain)) {
+		d_print("gain not available\n");
 		return;
 	}
-	if (parse_double(g, &gain) || parse_double(p, &peak)) {
-		d_print("could not parse gain (%s) or peak (%s)\n", g, p);
-		return;
+	if (isnan(peak)) {
+		d_print("peak not available, defaulting to 1\n");
+		peak = 1;
 	}
 	if (peak < 0.05) {
 		d_print("peak (%g) is too small\n", peak);
@@ -259,8 +381,10 @@ static void update_rg_scale(void)
 	scale = pow(10.0, db / 20.0);
 	replaygain_scale = scale;
 	limit = 1.0 / peak;
-	if (replaygain_limit && replaygain_scale > limit)
-		replaygain_scale = limit;
+	if (replaygain_limit && !isnan(peak)) {
+		if (replaygain_scale > limit)
+			replaygain_scale = limit;
+	}
 
 	d_print("gain = %f, peak = %f, db = %f, scale = %f, limit = %f, replaygain_scale = %f\n",
 			gain, peak, db, scale, limit, replaygain_scale);
@@ -278,50 +402,58 @@ static inline int get_next(struct track_info **ti)
 
 /* updating player status {{{ */
 
-static inline void file_changed(struct track_info *ti)
+static inline void _file_changed(struct track_info *ti)
 {
-	player_info_lock();
 	if (player_info.ti)
 		track_info_unref(player_info.ti);
 
 	player_info.ti = ti;
+	update_rg_scale();
+	player_info.metadata[0] = 0;
+	player_info.file_changed = 1;
+}
+
+static inline void file_changed(struct track_info *ti)
+{
+	player_info_lock();
 	if (ti) {
 		d_print("file: %s\n", ti->filename);
 	} else {
 		d_print("unloaded\n");
 	}
-	update_rg_scale();
-	player_info.metadata[0] = 0;
-	player_info.file_changed = 1;
+	_file_changed(ti);
 	player_info_unlock();
 }
 
 static inline void metadata_changed(void)
 {
+	struct keyval *comments;
+	int rc;
+
 	player_info_lock();
-	d_print("metadata changed: %s\n", ip_get_metadata(ip));
-	memcpy(player_info.metadata, ip_get_metadata(ip), 255 * 16 + 1);
+	if (ip_get_metadata(ip)) {
+		d_print("metadata changed: %s\n", ip_get_metadata(ip));
+		memcpy(player_info.metadata, ip_get_metadata(ip),
+		       255 * 16 + 1);
+	}
+
+	rc = ip_read_comments(ip, &comments);
+	if (!rc) {
+		if (player_info.ti->comments)
+			keyvals_free(player_info.ti->comments);
+		track_info_set_comments(player_info.ti, comments);
+	}
+
 	player_info.metadata_changed = 1;
 	player_info_unlock();
 }
 
-static inline void volume_update(int left, int right)
-{
-	if (player_info.vol_left == left && player_info.vol_right == right)
-		return;
-
-	player_info_lock();
-	player_info.vol_left = left;
-	player_info.vol_right = right;
-	player_info.vol_changed = 1;
-	player_info_unlock();
-}
-
 static void player_error(const char *msg)
 {
 	player_info_lock();
-	player_info.status = consumer_status;
+	player_info.status = (enum player_status)consumer_status;
 	player_info.pos = 0;
+	player_info.current_bitrate = -1;
 	player_info.buffer_fill = buffer_get_filled_chunks();
 	player_info.buffer_size = buffer_nr_chunks;
 	player_info.status_changed = 1;
@@ -333,7 +465,7 @@ static void player_error(const char *msg)
 	d_print("ERROR: '%s'\n", msg);
 }
 
-static void __FORMAT(2, 3) player_ip_error(int rc, const char *format, ...)
+static void CMUS_FORMAT(2, 3) player_ip_error(int rc, const char *format, ...)
 {
 	char buffer[1024];
 	va_list ap;
@@ -350,7 +482,7 @@ static void __FORMAT(2, 3) player_ip_error(int rc, const char *format, ...)
 	free(msg);
 }
 
-static void __FORMAT(2, 3) player_op_error(int rc, const char *format, ...)
+static void CMUS_FORMAT(2, 3) player_op_error(int rc, const char *format, ...)
 {
 	char buffer[1024];
 	va_list ap;
@@ -367,33 +499,10 @@ static void __FORMAT(2, 3) player_op_error(int rc, const char *format, ...)
 	free(msg);
 }
 
-/* FIXME: don't poll */
-static void mixer_check(void)
-{
-	static struct timeval old_t = { 0L, 0L };
-	struct timeval t;
-	long usec, sec;
-	int l, r;
-
-	gettimeofday(&t, NULL);
-	usec = t.tv_usec - old_t.tv_usec;
-	sec = t.tv_sec - old_t.tv_sec;
-	if (sec) {
-		/* multiplying sec with 1e6 can overflow */
-		usec += 1e6L;
-	}
-	if (usec < 300e3)
-		return;
-
-	old_t = t;
-	if (!op_get_volume(&l, &r))
-		volume_update(l, r);
-}
-
 /*
  * buffer-fill changed
  */
-static void __producer_buffer_fill_update(void)
+static void _producer_buffer_fill_update(void)
 {
 	int fill;
 
@@ -410,19 +519,25 @@ static void __producer_buffer_fill_update(void)
 /*
  * playing position changed
  */
-static void __consumer_position_update(void)
+static void _consumer_position_update(void)
 {
 	static unsigned int old_pos = -1;
 	unsigned int pos = 0;
-	
+	long bitrate;
+
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED)
 		pos = consumer_pos / buffer_second_size();
 	if (pos != old_pos) {
-/* 		d_print("\n"); */
 		old_pos = pos;
 
 		player_info_lock();
 		player_info.pos = pos;
+
+		if (show_current_bitrate) {
+			bitrate = ip_current_bitrate(ip);
+			if (bitrate != -1)
+				player_info.current_bitrate = bitrate;
+		}
 		player_info.position_changed = 1;
 		player_info_unlock();
 	}
@@ -431,7 +546,7 @@ static void __consumer_position_update(void)
 /*
  * something big happened (stopped/paused/unpaused...)
  */
-static void __player_status_changed(void)
+static void _player_status_changed(void)
 {
 	unsigned int pos = 0;
 
@@ -440,8 +555,9 @@ static void __player_status_changed(void)
 		pos = consumer_pos / buffer_second_size();
 
 	player_info_lock();
-	player_info.status = consumer_status;
+	player_info.status = (enum player_status)consumer_status;
 	player_info.pos = pos;
+	player_info.current_bitrate = -1;
 	player_info.buffer_fill = buffer_get_filled_chunks();
 	player_info.buffer_size = buffer_nr_chunks;
 	player_info.status_changed = 1;
@@ -450,7 +566,7 @@ static void __player_status_changed(void)
 
 /* updating player status }}} */
 
-static void __prebuffer(void)
+static void _prebuffer(void)
 {
 	int limit_chunks;
 
@@ -494,7 +610,7 @@ static void __prebuffer(void)
 		/* buffer_fill with 0 count marks current chunk filled */
 		buffer_fill(nr_read);
 
-		__producer_buffer_fill_update();
+		_producer_buffer_fill_update();
 		if (nr_read == 0) {
 			/* EOF */
 			break;
@@ -504,7 +620,13 @@ static void __prebuffer(void)
 
 /* setting producer status {{{ */
 
-static void __producer_play(void)
+static void _producer_status_update(enum producer_status status)
+{
+	producer_status =  status;
+	pthread_cond_broadcast(&producer_playing);
+}
+
+static void _producer_play(void)
 {
 	if (producer_status == PS_UNLOADED) {
 		struct track_info *ti;
@@ -521,7 +643,7 @@ static void __producer_play(void)
 				file_changed(NULL);
 			} else {
 				ip_setup(ip);
-				producer_status = PS_PLAYING;
+				_producer_status_update(PS_PLAYING);
 				file_changed(ti);
 			}
 		}
@@ -536,48 +658,48 @@ static void __producer_play(void)
 		if (rc) {
 			player_ip_error(rc, "opening file `%s'", ip_get_filename(ip));
 			ip_delete(ip);
-			producer_status = PS_UNLOADED;
+			_producer_status_update(PS_UNLOADED);
 		} else {
 			ip_setup(ip);
-			producer_status = PS_PLAYING;
+			_producer_status_update(PS_PLAYING);
 		}
 	} else if (producer_status == PS_PAUSED) {
-		producer_status = PS_PLAYING;
+		_producer_status_update(PS_PLAYING);
 	}
 }
 
-static void __producer_stop(void)
+static void _producer_stop(void)
 {
 	if (producer_status == PS_PLAYING || producer_status == PS_PAUSED) {
 		ip_close(ip);
-		producer_status = PS_STOPPED;
+		_producer_status_update(PS_STOPPED);
 		reset_buffer();
 	}
 }
 
-static void __producer_unload(void)
+static void _producer_unload(void)
 {
-	__producer_stop();
+	_producer_stop();
 	if (producer_status == PS_STOPPED) {
 		ip_delete(ip);
-		producer_status = PS_UNLOADED;
+		_producer_status_update(PS_UNLOADED);
 	}
 }
 
-static void __producer_pause(void)
+static void _producer_pause(void)
 {
 	if (producer_status == PS_PLAYING) {
-		producer_status = PS_PAUSED;
+		_producer_status_update(PS_PAUSED);
 	} else if (producer_status == PS_PAUSED) {
-		producer_status = PS_PLAYING;
+		_producer_status_update(PS_PLAYING);
 	}
 }
 
-static void __producer_set_file(struct track_info *ti)
+static void _producer_set_file(struct track_info *ti)
 {
-	__producer_unload();
+	_producer_unload();
 	ip = ip_new(ti->filename);
-	producer_status = PS_STOPPED;
+	_producer_status_update(PS_STOPPED);
 	file_changed(ti);
 }
 
@@ -585,121 +707,145 @@ static void __producer_set_file(struct track_info *ti)
 
 /* setting consumer status {{{ */
 
-static void __consumer_play(void)
+static void _consumer_status_update(enum consumer_status status)
+{
+	consumer_status = status;
+	pthread_cond_broadcast(&consumer_playing);
+}
+
+static void _consumer_play(void)
 {
 	if (consumer_status == CS_PLAYING) {
 		op_drop();
 	} else if (consumer_status == CS_STOPPED) {
 		int rc;
 
-		set_buffer_sf(ip_get_sf(ip));
-		rc = op_open(buffer_sf);
+		set_buffer_sf();
+		rc = op_open(buffer_sf, buffer_channel_map);
 		if (rc) {
 			player_op_error(rc, "opening audio device");
 		} else {
-			consumer_status = CS_PLAYING;
+			_consumer_status_update(CS_PLAYING);
 		}
 	} else if (consumer_status == CS_PAUSED) {
 		op_unpause();
-		consumer_status = CS_PLAYING;
+		_consumer_status_update(CS_PLAYING);
 	}
 }
 
-static void __consumer_drain_and_stop(void)
+static void _consumer_drain_and_stop(void)
 {
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED) {
 		op_close();
-		consumer_status = CS_STOPPED;
+		_consumer_status_update(CS_STOPPED);
 	}
 }
 
-static void __consumer_stop(void)
+static void _consumer_stop(void)
 {
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED) {
 		op_drop();
 		op_close();
-		consumer_status = CS_STOPPED;
+		_consumer_status_update(CS_STOPPED);
 	}
 }
 
-static void __consumer_pause(void)
+static void _consumer_pause(void)
 {
 	if (consumer_status == CS_PLAYING) {
 		op_pause();
-		consumer_status = CS_PAUSED;
+		_consumer_status_update(CS_PAUSED);
 	} else if (consumer_status == CS_PAUSED) {
 		op_unpause();
-		consumer_status = CS_PLAYING;
+		_consumer_status_update(CS_PLAYING);
 	}
 }
 
 /* setting consumer status }}} */
 
-static int change_sf(sample_format_t sf, int drop)
+
+static int change_sf(int drop)
 {
 	int old_sf = buffer_sf;
+	CHANNEL_MAP(old_channel_map);
+	channel_map_copy(old_channel_map, buffer_channel_map);
 
-	set_buffer_sf(sf);
-	if (buffer_sf != old_sf) {
+	set_buffer_sf();
+	if (buffer_sf != old_sf || !channel_map_equal(buffer_channel_map, old_channel_map, sf_get_channels(buffer_sf))) {
 		/* reopen */
 		int rc;
 
 		if (drop)
 			op_drop();
 		op_close();
-		rc = op_open(buffer_sf);
+		rc = op_open(buffer_sf, buffer_channel_map);
 		if (rc) {
 			player_op_error(rc, "opening audio device");
-			consumer_status = CS_STOPPED;
-			__producer_stop();
+			_consumer_status_update(CS_STOPPED);
+			_producer_stop();
 			return rc;
 		}
 	} else if (consumer_status == CS_PAUSED) {
 		op_drop();
 		op_unpause();
 	}
-	consumer_status = CS_PLAYING;
+	_consumer_status_update(CS_PLAYING);
 	return 0;
 }
 
-static void __consumer_handle_eof(void)
+static void _consumer_handle_eof(void)
 {
 	struct track_info *ti;
 
 	if (ip_is_remote(ip)) {
-		__producer_stop();
-		__consumer_drain_and_stop();
+		_producer_stop();
+		_consumer_drain_and_stop();
 		player_error("lost connection");
 		return;
 	}
 
+	if (player_info.ti)
+		player_info.ti->play_count++;
+
+	if (player_repeat_current) {
+		if (player_cont) {
+			ip_seek(ip, 0);
+			reset_buffer();
+		} else {
+			_producer_stop();
+			_consumer_drain_and_stop();
+			_player_status_changed();
+		}
+		return;
+	}
+
 	if (get_next(&ti) == 0) {
-		__producer_unload();
+		_producer_unload();
 		ip = ip_new(ti->filename);
-		producer_status = PS_STOPPED;
+		_producer_status_update(PS_STOPPED);
 		/* PS_STOPPED, CS_PLAYING */
 		if (player_cont) {
-			__producer_play();
+			_producer_play();
 			if (producer_status == PS_UNLOADED) {
-				__consumer_stop();
+				_consumer_stop();
 				track_info_unref(ti);
 				file_changed(NULL);
 			} else {
 				/* PS_PLAYING */
 				file_changed(ti);
-				if (!change_sf(ip_get_sf(ip), 0))
-					__prebuffer();
+				if (!change_sf(0))
+					_prebuffer();
 			}
 		} else {
-			__consumer_drain_and_stop();
+			_consumer_drain_and_stop();
 			file_changed(ti);
 		}
 	} else {
-		__producer_unload();
-		__consumer_drain_and_stop();
+		_producer_unload();
+		_consumer_drain_and_stop();
 		file_changed(NULL);
 	}
-	__player_status_changed();
+	_player_status_changed();
 }
 
 static void *consumer_loop(void *arg)
@@ -714,26 +860,28 @@ static void *consumer_loop(void *arg)
 			break;
 
 		if (consumer_status == CS_PAUSED || consumer_status == CS_STOPPED) {
-			mixer_check();
+			pthread_cond_wait(&consumer_playing, &consumer_mutex);
 			consumer_unlock();
-			ms_sleep(50);
 			continue;
 		}
 		space = op_buffer_space();
-		if (space == -1) {
-			/* busy */
-			__consumer_position_update();
+		if (space < 0) {
+			d_print("op_buffer_space returned %d %s\n", space,
+					space == -1 ? strerror(errno) : "");
+
+			/* try to reopen */
+			op_close();
+			_consumer_status_update(CS_STOPPED);
+			_consumer_play();
+
 			consumer_unlock();
-			ms_sleep(50);
 			continue;
 		}
 /* 		d_print("BS: %6d %3d\n", space, space * 1000 / (44100 * 2 * 2)); */
 
 		while (1) {
-			/* 25 ms is 4410 B */
-			if (space < 4096) {
-				__consumer_position_update();
-				mixer_check();
+			if (space == 0) {
+				_consumer_position_update();
 				consumer_unlock();
 				ms_sleep(25);
 				break;
@@ -752,14 +900,14 @@ static void *consumer_loop(void *arg)
 					/* OK. now it's safe to check if we are at EOF */
 					if (ip_eof(ip)) {
 						/* EOF */
-						__consumer_handle_eof();
+						_consumer_handle_eof();
 						producer_unlock();
 						consumer_unlock();
 						break;
 					} else {
 						/* possible underrun */
 						producer_unlock();
-						__consumer_position_update();
+						_consumer_position_update();
 						consumer_unlock();
 /* 						d_print("possible underrun\n"); */
 						ms_sleep(10);
@@ -773,7 +921,7 @@ static void *consumer_loop(void *arg)
 			if (size > space)
 				size = space;
 			if (soft_vol || replaygain)
-				scale_samples(rpos, &size);
+				scale_samples(rpos, (unsigned int *)&size);
 			rc = op_write(rpos, size);
 			if (rc < 0) {
 				d_print("op_write returned %d %s\n", rc,
@@ -781,8 +929,8 @@ static void *consumer_loop(void *arg)
 
 				/* try to reopen */
 				op_close();
-				consumer_status = CS_STOPPED;
-				__consumer_play();
+				_consumer_status_update(CS_STOPPED);
+				_consumer_play();
 
 				consumer_unlock();
 				break;
@@ -792,7 +940,7 @@ static void *consumer_loop(void *arg)
 			space -= rc;
 		}
 	}
-	__consumer_stop();
+	_consumer_stop();
 	consumer_unlock();
 	return NULL;
 }
@@ -815,8 +963,8 @@ static void *producer_loop(void *arg)
 		if (producer_status == PS_UNLOADED ||
 		    producer_status == PS_PAUSED ||
 		    producer_status == PS_STOPPED || ip_eof(ip)) {
+			pthread_cond_wait(&producer_playing, &producer_mutex);
 			producer_unlock();
-			ms_sleep(50);
 			continue;
 		}
 		for (i = 0; ; i++) {
@@ -857,23 +1005,17 @@ static void *producer_loop(void *arg)
 				break;
 			}
 		}
-		__producer_buffer_fill_update();
+		_producer_buffer_fill_update();
 	}
-	__producer_unload();
+	_producer_unload();
 	producer_unlock();
 	return NULL;
 }
 
-void player_load_plugins(void)
-{
-	ip_load_plugins();
-	op_load_plugins();
-}
-
 void player_init(const struct player_callbacks *callbacks)
 {
 	int rc;
-#if defined(__linux__) || defined(__FreeBSD__)
+#ifdef REALTIME_SCHEDULING
 	pthread_attr_t attr;
 #endif
 	pthread_attr_t *attrp = NULL;
@@ -886,7 +1028,7 @@ void player_init(const struct player_callbacks *callbacks)
 
 	player_cbs = callbacks;
 
-#if defined(__linux__) || defined(__FreeBSD__)
+#ifdef REALTIME_SCHEDULING
 	rc = pthread_attr_init(&attr);
 	BUG_ON(rc);
 	rc = pthread_attr_setschedpolicy(&attr, SCHED_RR);
@@ -916,7 +1058,7 @@ void player_init(const struct player_callbacks *callbacks)
 
 	/* update player_info.cont etc. */
 	player_lock();
-	__player_status_changed();
+	_player_status_changed();
 	player_unlock();
 }
 
@@ -926,23 +1068,24 @@ void player_exit(void)
 
 	player_lock();
 	consumer_running = 0;
+	pthread_cond_broadcast(&consumer_playing);
 	producer_running = 0;
+	pthread_cond_broadcast(&producer_playing);
 	player_unlock();
-	
+
 	rc = pthread_join(consumer_thread, NULL);
 	BUG_ON(rc);
 	rc = pthread_join(producer_thread, NULL);
 	BUG_ON(rc);
-
-	op_exit_plugins();
+	buffer_free();
 }
 
 void player_stop(void)
 {
 	player_lock();
-	__consumer_stop();
-	__producer_stop();
-	__player_status_changed();
+	_consumer_stop();
+	_producer_stop();
+	_player_status_changed();
 	player_unlock();
 }
 
@@ -957,119 +1100,129 @@ void player_play(void)
 		return;
 	}
 	prebuffer = consumer_status == CS_STOPPED;
-	__producer_play();
+	_producer_play();
 	if (producer_status == PS_PLAYING) {
-		__consumer_play();
+		_consumer_play();
 		if (consumer_status != CS_PLAYING)
-			__producer_stop();
+			_producer_stop();
 	} else {
-		__consumer_stop();
+		_consumer_stop();
 	}
-	__player_status_changed();
+	_player_status_changed();
 	if (consumer_status == CS_PLAYING && prebuffer)
-		__prebuffer();
+		_prebuffer();
 	player_unlock();
 }
 
 void player_pause(void)
 {
+	if (ip && ip_is_remote(ip) && consumer_status == CS_PLAYING) {
+		/* pausing not allowed */
+		player_stop();
+		return;
+	}
 	player_lock();
 
 	if (consumer_status == CS_STOPPED) {
-		__producer_play();
+		_producer_play();
 		if (producer_status == PS_PLAYING) {
-			__consumer_play();
+			_consumer_play();
 			if (consumer_status != CS_PLAYING)
-				__producer_stop();
+				_producer_stop();
 		}
-		__player_status_changed();
+		_player_status_changed();
 		if (consumer_status == CS_PLAYING)
-			__prebuffer();
+			_prebuffer();
 		player_unlock();
 		return;
 	}
 
-	if (ip && ip_is_remote(ip)) {
-		/* pausing not allowed */
-		player_unlock();
-		return;
-	}
-	__producer_pause();
-	__consumer_pause();
-	__player_status_changed();
+	_producer_pause();
+	_consumer_pause();
+	_player_status_changed();
 	player_unlock();
 }
 
 void player_set_file(struct track_info *ti)
 {
 	player_lock();
-	__producer_set_file(ti);
+	_producer_set_file(ti);
 	if (producer_status == PS_UNLOADED) {
-		__consumer_stop();
+		_consumer_stop();
 		goto out;
 	}
 
 	/* PS_STOPPED */
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED) {
-		__producer_play();
+		op_drop();
+		_producer_play();
 		if (producer_status == PS_UNLOADED) {
-			__consumer_stop();
+			_consumer_stop();
 			goto out;
 		}
-		change_sf(ip_get_sf(ip), 1);
+		change_sf(1);
 	}
 out:
-	__player_status_changed();
+	_player_status_changed();
 	if (producer_status == PS_PLAYING)
-		__prebuffer();
+		_prebuffer();
 	player_unlock();
 }
 
 void player_play_file(struct track_info *ti)
 {
 	player_lock();
-	__producer_set_file(ti);
+	_producer_set_file(ti);
 	if (producer_status == PS_UNLOADED) {
-		__consumer_stop();
+		_consumer_stop();
 		goto out;
 	}
 
 	/* PS_STOPPED */
-	__producer_play();
+	_producer_play();
 
 	/* PS_UNLOADED,PS_PLAYING */
 	if (producer_status == PS_UNLOADED) {
-		__consumer_stop();
+		_consumer_stop();
 		goto out;
 	}
 
 	/* PS_PLAYING */
 	if (consumer_status == CS_STOPPED) {
-		__consumer_play();
+		_consumer_play();
 		if (consumer_status == CS_STOPPED)
-			__producer_stop();
+			_producer_stop();
 	} else {
-		change_sf(ip_get_sf(ip), 1);
+		op_drop();
+		change_sf(1);
 	}
 out:
-	__player_status_changed();
+	_player_status_changed();
 	if (producer_status == PS_PLAYING)
-		__prebuffer();
+		_prebuffer();
 	player_unlock();
 }
 
-void player_seek(double offset, int relative)
+void player_file_changed(struct track_info *ti)
 {
+	_file_changed(ti);
+}
+
+void player_seek(double offset, int relative, int start_playing)
+{
+	int stopped = 0;
 	player_lock();
 	if (consumer_status == CS_STOPPED) {
-		__producer_play();
+		stopped = 1;
+		_producer_play();
 		if (producer_status == PS_PLAYING) {
-			__consumer_play();
+			_consumer_play();
 			if (consumer_status != CS_PLAYING) {
-				__producer_stop();
+				_producer_stop();
 				player_unlock();
 				return;
-			}
+			} else
+				_player_status_changed();
 		}
 	}
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED) {
@@ -1117,13 +1270,19 @@ void player_seek(double offset, int relative)
 /* 		d_print("seeking %g/%g (%g from eof)\n", new_pos, duration, duration - new_pos); */
 		rc = ip_seek(ip, new_pos);
 		if (rc == 0) {
-/* 			d_print("doing op_drop after seek\n"); */
+			d_print("doing op_drop after seek\n");
 			op_drop();
 			reset_buffer();
 			consumer_pos = new_pos * buffer_second_size();
 			scale_pos = consumer_pos;
-			__consumer_position_update();
+			_consumer_position_update();
+			if (stopped && !start_playing) {
+				_producer_pause();
+				_consumer_pause();
+				_player_status_changed();
+			}
 		} else {
+			player_ip_error(rc, "seeking in file %s", ip_get_filename(ip));
 			d_print("error: ip_seek returned %d\n", rc);
 		}
 	}
@@ -1135,7 +1294,7 @@ void player_seek(double offset, int relative)
  */
 void player_set_op(const char *name)
 {
-	int rc, l, r;
+	int rc;
 
 	player_lock();
 
@@ -1155,20 +1314,23 @@ void player_set_op(const char *name)
 		rc = op_select_any();
 	}
 	if (rc) {
-		consumer_status = CS_STOPPED;
+		_consumer_status_update(CS_STOPPED);
 
-		__producer_stop();
-		player_op_error(rc, "selecting output plugin '%s'", name);
+		_producer_stop();
+		if (name)
+			player_op_error(rc, "selecting output plugin '%s'", name);
+		else
+			player_op_error(rc, "selecting any output plugin");
 		player_unlock();
 		return;
 	}
 
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED) {
-		set_buffer_sf(ip_get_sf(ip));
-		rc = op_open(buffer_sf);
+		set_buffer_sf();
+		rc = op_open(buffer_sf, buffer_channel_map);
 		if (rc) {
-			consumer_status = CS_STOPPED;
-			__producer_stop();
+			_consumer_status_update(CS_STOPPED);
+			_producer_stop();
 			player_op_error(rc, "opening audio device");
 			player_unlock();
 			return;
@@ -1177,32 +1339,19 @@ void player_set_op(const char *name)
 			op_pause();
 	}
 
-	if (!op_get_volume(&l, &r))
-		volume_update(l, r);
-
 	player_unlock();
 }
 
-char *player_get_op(void)
-{
-	return op_get_current();
-}
-
 void player_set_buffer_chunks(unsigned int nr_chunks)
 {
-	if (nr_chunks < 3)
-		nr_chunks = 3;
-	if (nr_chunks > 30)
-		nr_chunks = 30;
-
 	player_lock();
-	__producer_stop();
-	__consumer_stop();
+	_producer_stop();
+	_consumer_stop();
 
 	buffer_nr_chunks = nr_chunks;
 	buffer_init();
 
-	__player_status_changed();
+	_player_status_changed();
 	player_unlock();
 }
 
@@ -1211,69 +1360,21 @@ int player_get_buffer_chunks(void)
 	return buffer_nr_chunks;
 }
 
-int player_get_fileinfo(const char *filename, int *duration,
-		struct keyval **comments)
+void player_set_soft_volume(int l, int r)
 {
-	struct input_plugin *plug;
-	int rc;
-
-	*comments = NULL;
-	*duration = -1;
-	plug = ip_new(filename);
-	if (ip_is_remote(plug)) {
-		*comments = xnew0(struct keyval, 1);
-		ip_delete(plug);
-		return 0;
-	}
-	rc = ip_open(plug);
-	if (rc) {
-		int save = errno;
-
-		ip_delete(plug);
-		errno = save;
-		if (rc != -1)
-			rc = -PLAYER_ERROR_NOT_SUPPORTED;
-		return rc;
-	}
-	*duration = ip_duration(plug);
-	rc = ip_read_comments(plug, comments);
-	ip_delete(plug);
-	return rc;
-}
-
-int player_get_volume(int *left, int *right)
-{
-	int rc;
-
 	consumer_lock();
-	rc = op_get_volume(left, right);
+	soft_vol_l = l;
+	soft_vol_r = r;
 	consumer_unlock();
-	return rc;
-}
-
-int player_set_volume(int left, int right)
-{
-	int rc;
-
-	consumer_lock();
-	rc = op_set_volume(left, right);
-	if (!rc)
-		volume_update(left, right);
-	consumer_unlock();
-	return rc;
 }
 
 void player_set_soft_vol(int soft)
 {
-	int l, r;
-
 	consumer_lock();
 	/* don't mess with scale_pos if soft_vol or replaygain is already enabled */
 	if (!soft_vol && !replaygain)
 		scale_pos = consumer_pos;
-	op_set_soft_vol(soft);
-	if (!op_get_volume(&l, &r))
-		volume_update(l, r);
+	soft_vol = soft;
 	consumer_unlock();
 }
 
@@ -1315,43 +1416,3 @@ void player_set_rg_preamp(double db)
 
 	player_unlock();
 }
-
-int player_set_op_option(unsigned int id, const char *val)
-{
-	int rc;
-
-	player_lock();
-	__consumer_stop();
-	__producer_stop();
-	rc = op_set_option(id, val);
-	__player_status_changed();
-	player_unlock();
-	return rc;
-}
-
-int player_get_op_option(unsigned int id, char **val)
-{
-	int rc;
-
-	player_lock();
-	rc = op_get_option(id, val);
-	player_unlock();
-	return rc;
-}
-
-int player_for_each_op_option(void (*callback)(unsigned int id, const char *key))
-{
-	player_lock();
-	__consumer_stop();
-	__producer_stop();
-	op_for_each_option(callback);
-	__player_status_changed();
-	player_unlock();
-	return 0;
-}
-
-void player_dump_plugins(void)
-{
-	ip_dump_plugins();
-	op_dump_plugins();
-}
