@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,11 +11,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "search_mode.h"
@@ -26,15 +25,11 @@
 #include "xstrjoin.h"
 #include "misc.h"
 #include "lib.h"
+#include "command_mode.h"
+#include "keys.h"
 
 #include <ctype.h>
 
-#if defined(__sun__)
-#include <ncurses.h>
-#else
-#include <curses.h>
-#endif
-
 /* this is set in ui_curses.c */
 enum search_direction search_direction = SEARCH_FORWARD;
 
@@ -45,7 +40,6 @@ enum search_direction search_direction = SEARCH_FORWARD;
 char *search_str = NULL;
 int search_restricted = 0;
 
-static int search_found = 0;
 static struct history search_history;
 static char *search_history_filename;
 static char *history_search_text = NULL;
@@ -115,12 +109,35 @@ static void delete(void)
 	cmdline_delete_ch();
 	parse_line(&text, &search_restricted);
 	if (text[0])
-		search_found = search(searchable, text, search_direction, 0);
+		search(searchable, text, search_direction, 0);
 
 	/* restore old value */
 	search_restricted = restricted;
 }
 
+void search_text(const char *text, int restricted, int beginning)
+{
+	if (text[0] == 0) {
+		/* cmdline is "/", "?", "//" or "??" */
+		if (search_str) {
+			/* use old search string */
+			search_restricted = restricted;
+			if (!search_next(searchable, search_str, search_direction))
+				search_not_found();
+		}
+	} else {
+		/* set new search string and add it to the history */
+		free(search_str);
+		search_str = xstrdup(text);
+		history_add_line(&search_history, text);
+
+		/* search not yet done if up or down arrow was pressed */
+		search_restricted = restricted;
+		if (!search(searchable, search_str, search_direction, beginning))
+			search_not_found();
+	}
+}
+
 void search_mode_ch(uchar ch)
 {
 	const char *text;
@@ -154,26 +171,8 @@ void search_mode_ch(uchar ch)
 		break;
 	case 0x0A:
 		parse_line(&text, &restricted);
-		if (text[0] == 0) {
-			/* cmdline is "/", "?", "//" or "??" */
-			if (search_str) {
-				/* use old search string */
-				search_restricted = restricted;
-				search_found = search_next(searchable, search_str, search_direction);
-			}
-		} else {
-			/* set new search string and add it to the history */
-			free(search_str);
-			search_str = xstrdup(text);
-			history_add_line(&search_history, text);
-
-			/* search not yet done if up or down arrow was pressed */
-			search_restricted = restricted;
-			search_found = search(searchable, search_str, search_direction, 0);
-		}
+		search_text(text, restricted, 0);
 		cmdline_clear();
-		if (!search_found)
-			search_not_found();
 		input_mode = NORMAL_MODE;
 		break;
 	case 0x0B:
@@ -182,6 +181,10 @@ void search_mode_ch(uchar ch)
 	case 0x15:
 		cmdline_backspace_to_bol();
 		break;
+	case 0x17: // ^W
+		cmdline_backward_delete_word(cmdline_word_delimiters);
+		break;
+	case 0x08: // ^H
 	case 127:
 		backspace();
 		break;
@@ -201,7 +204,7 @@ void search_mode_ch(uchar ch)
 
 			cmdline_insert_ch(ch);
 			parse_line(&text, &search_restricted);
-			search_found = search(searchable, text, search_direction, beginning);
+			search(searchable, text, search_direction, beginning);
 
 			/* restore old value */
 			search_restricted = restricted;
@@ -211,6 +214,26 @@ void search_mode_ch(uchar ch)
 	reset_history_search();
 }
 
+void search_mode_escape(int c)
+{
+	switch (c) {
+	case 98:
+		cmdline_backward_word(cmdline_filename_delimiters);
+		break;
+	case 100:
+		cmdline_delete_word(cmdline_filename_delimiters);
+		break;
+	case 102:
+		cmdline_forward_word(cmdline_filename_delimiters);
+		break;
+	case 127:
+	case KEY_BACKSPACE:
+		cmdline_backward_delete_word(cmdline_filename_delimiters);
+		break;
+	}
+	reset_history_search();
+}
+
 void search_mode_key(int key)
 {
 	const char *text;
@@ -260,6 +283,35 @@ void search_mode_key(int key)
 	reset_history_search();
 }
 
+void search_mode_mouse(MEVENT *event)
+{
+	if ((event->bstate & BUTTON1_PRESSED) || (event->bstate & BUTTON3_PRESSED)) {
+		const char *text;
+		int restricted;
+		if (event->y <= window_get_nr_rows(current_win()) + 2) {
+			parse_line(&text, &restricted);
+			if (text[0]) {
+				history_add_line(&search_history, text);
+				cmdline_clear();
+			}
+			input_mode = NORMAL_MODE;
+			normal_mode_mouse(event);
+			return;
+		}
+		if (event->x == 0)
+			return;
+		int i = event->x > cmdline.clen ? cmdline.clen : event->x - 1;
+		while (i < cmdline.cpos)
+			cmdline_move_left();
+		while (i > cmdline.cpos)
+			cmdline_move_right();
+	} else if (event->bstate & BUTTON4_PRESSED) {
+		search_mode_key(KEY_UP);
+	} else if (event->bstate & BUTTON5_PRESSED) {
+		search_mode_key(KEY_DOWN);
+	}
+}
+
 void search_mode_init(void)
 {
 	search_history_filename = xstrjoin(cmus_config_dir, "/search-history");
@@ -269,5 +321,6 @@ void search_mode_init(void)
 void search_mode_exit(void)
 {
 	history_save(&search_history);
+	history_free(&search_history);
 	free(search_history_filename);
 }
