@@ -1,5 +1,5 @@
 /*
- * Copyright 2008-2011 Various Authors
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -16,6 +16,7 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#include "convert.h"
 #include "ui_curses.h"
 #include "cmdline.h"
 #include "search_mode.h"
@@ -33,7 +34,6 @@
 #include "xmalloc.h"
 #include "xstrjoin.h"
 #include "window.h"
-#include "format_print.h"
 #include "comment.h"
 #include "misc.h"
 #include "prog.h"
@@ -54,6 +54,7 @@
 #endif
 
 #include <unistd.h>
+#include <fcntl.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
@@ -69,6 +70,7 @@
 #include <signal.h>
 #include <stdarg.h>
 #include <math.h>
+#include <sys/time.h>
 
 #if defined(__sun__) || defined(__CYGWIN__)
 /* TIOCGWINSZ */
@@ -84,7 +86,7 @@ char *tgoto(const char *cap, int col, int row);
 
 /* globals. documented in ui_curses.h */
 
-int cmus_running = 1;
+sig_atomic_t cmus_running = 1;
 int ui_initialized = 0;
 enum ui_input_mode input_mode = NORMAL_MODE;
 int cur_view = TREE_VIEW;
@@ -105,6 +107,9 @@ static char *lib_autosave_filename;
 static char *pl_autosave_filename;
 static char *play_queue_autosave_filename;
 
+static int notify_in;
+static int notify_out;
+
 /* shown error message and time stamp
  * error is cleared if it is older than 3s and key was pressed
  */
@@ -118,7 +123,7 @@ static char *server_address = NULL;
 
 static char print_buffer[512];
 
-/* destination buffer for utf8_encode and utf8_decode */
+/* destination buffer for utf8_encode_to_buf and utf8_decode */
 static char conv_buffer[512];
 
 /* one character can take up to 4 bytes in UTF-8 */
@@ -143,6 +148,8 @@ static int show_cursor;
 static int cursor_x;
 static int cursor_y;
 
+static char *title_buf = NULL;
+
 enum {
 	CURSED_WIN,
 	CURSED_WIN_CUR,
@@ -211,6 +218,28 @@ static unsigned char cursed_to_fg_idx[NR_CURSED] = {
 	COLOR_INFO
 };
 
+static unsigned char cursed_to_attr_idx[NR_CURSED] = {
+	COLOR_WIN_ATTR,
+	COLOR_WIN_ATTR,
+	COLOR_WIN_INACTIVE_SEL_ATTR,
+	COLOR_WIN_INACTIVE_CUR_SEL_ATTR,
+
+	COLOR_WIN_ATTR,
+	COLOR_WIN_ATTR,
+	COLOR_WIN_SEL_ATTR,
+	COLOR_WIN_CUR_SEL_ATTR,
+
+	COLOR_WIN_ATTR,
+	COLOR_WIN_TITLE_ATTR,
+	COLOR_CMDLINE_ATTR,
+	COLOR_STATUSLINE_ATTR,
+
+	COLOR_TITLELINE_ATTR,
+	COLOR_WIN_ATTR,
+	COLOR_CMDLINE_ATTR,
+	COLOR_CMDLINE_ATTR
+};
+
 /* index is CURSED_*, value is fucking color pair */
 static int pairs[NR_CURSED];
 
@@ -221,11 +250,14 @@ enum {
 	TF_DISC,
 	TF_TRACK,
 	TF_TITLE,
+	TF_PLAY_COUNT,
 	TF_YEAR,
+	TF_MAX_YEAR,
 	TF_ORIGINALYEAR,
 	TF_GENRE,
 	TF_COMMENT,
 	TF_DURATION,
+	TF_DURATION_SEC,
 	TF_BITRATE,
 	TF_CODEC,
 	TF_CODEC_PROFILE,
@@ -249,6 +281,21 @@ enum {
 	TF_PART,
 	TF_SUBTITLE,
 	TF_MEDIA,
+	TF_VA,
+	TF_STATUS,
+	TF_POSITION,
+	TF_POSITION_SEC,
+	TF_TOTAL,
+	TF_VOLUME,
+	TF_LVOLUME,
+	TF_RVOLUME,
+	TF_BUFFER,
+	TF_REPEAT,
+	TF_CONTINUE,
+	TF_FOLLOW,
+	TF_SHUFFLE,
+	TF_PLAYLISTMODE,
+
 	NR_TFS
 };
 
@@ -259,11 +306,14 @@ static struct format_option track_fopts[NR_TFS + 1] = {
 	DEF_FO_INT('D', "discnumber", 1),
 	DEF_FO_INT('n', "tracknumber", 1),
 	DEF_FO_STR('t', "title", 0),
-	DEF_FO_STR('y', "date", 1),
-	DEF_FO_STR('\0', "originaldate", 1),
+	DEF_FO_INT('X', "play_count", 0),
+	DEF_FO_INT('y', "date", 1),
+	DEF_FO_INT('\0', "maxdate", 1),
+	DEF_FO_INT('\0', "originaldate", 1),
 	DEF_FO_STR('g', "genre", 0),
 	DEF_FO_STR('c', "comment", 0),
 	DEF_FO_TIME('d', "duration", 0),
+	DEF_FO_INT('\0', "duration_sec", 1),
 	DEF_FO_INT('\0', "bitrate", 0),
 	DEF_FO_STR('\0', "codec", 0),
 	DEF_FO_STR('\0', "codec_profile", 0),
@@ -287,49 +337,36 @@ static struct format_option track_fopts[NR_TFS + 1] = {
 	DEF_FO_STR('\0', "part", 0),
 	DEF_FO_STR('\0', "subtitle", 0),
 	DEF_FO_STR('\0', "media", 0),
+	DEF_FO_INT('\0', "va", 0),
+	DEF_FO_STR('\0', "status", 0),
+	DEF_FO_TIME('\0', "position", 0),
+	DEF_FO_INT('\0', "position_sec", 1),
+	DEF_FO_TIME('\0', "total", 0),
+	DEF_FO_INT('\0', "volume", 1),
+	DEF_FO_INT('\0', "lvolume", 1),
+	DEF_FO_INT('\0', "rvolume", 1),
+	DEF_FO_INT('\0', "buffer", 1),
+	DEF_FO_STR('\0', "repeat", 0),
+	DEF_FO_STR('\0', "continue", 0),
+	DEF_FO_STR('\0', "follow", 0),
+	DEF_FO_STR('\0', "shuffle", 0),
+	DEF_FO_STR('\0', "playlist_mode", 0),
 	DEF_FO_END
 };
 
-enum {
-	SF_STATUS,
-	SF_POSITION,
-	SF_DURATION,
-	SF_TOTAL,
-	SF_VOLUME,
-	SF_LVOLUME,
-	SF_RVOLUME,
-	SF_BUFFER,
-	SF_REPEAT,
-	SF_CONTINUE,
-	SF_SHUFFLE,
-	SF_PLAYLISTMODE,
-	SF_BITRATE,
-	NR_SFS
-};
+static int last_mevent = 0;
 
-static struct format_option status_fopts[NR_SFS + 1] = {
-	DEF_FO_STR('s', NULL, 0),
-	DEF_FO_TIME('p', NULL, 0),
-	DEF_FO_TIME('d', NULL, 0),
-	DEF_FO_TIME('t', NULL, 0),
-	DEF_FO_INT('v', NULL, 0),
-	DEF_FO_INT('l', NULL, 0),
-	DEF_FO_INT('r', NULL, 0),
-	DEF_FO_INT('b', NULL, 0),
-	DEF_FO_STR('R', NULL, 0),
-	DEF_FO_STR('C', NULL, 0),
-	DEF_FO_STR('S', NULL, 0),
-	DEF_FO_STR('L', NULL, 0),
-	DEF_FO_INT('B', NULL, 0),
-	DEF_FO_END
-};
+int get_track_win_x(void)
+{
+	return track_win_x;
+}
 
 int track_format_valid(const char *format)
 {
 	return format_valid(format, track_fopts);
 }
 
-static void utf8_encode(const char *buffer)
+static void utf8_encode_to_buf(const char *buffer)
 {
 	int n;
 #ifdef HAVE_ICONV
@@ -492,52 +529,6 @@ static void sprint_ascii(int row, int col, const char *str, int len)
 	(void) mvaddstr(row, col, print_buffer);
 }
 
-static void print_tree(struct window *win, int row, struct iter *iter)
-{
-	const char *str;
-	struct artist *artist;
-	struct album *album;
-	struct iter sel;
-	int current, selected, active, pos;
-
-	artist = iter_to_artist(iter);
-	album = iter_to_album(iter);
-	current = 0;
-	if (lib_cur_track) {
-		if (album) {
-			current = CUR_ALBUM == album;
-		} else {
-			current = CUR_ARTIST == artist;
-		}
-	}
-	window_get_sel(win, &sel);
-	selected = iters_equal(iter, &sel);
-	active = lib_cur_win == lib_tree_win;
-	bkgdset(pairs[(active << 2) | (selected << 1) | current]);
-
-	if (active && selected) {
-		cursor_x = 0;
-		cursor_y = 1 + row;
-	}
-
-	pos = 0;
-	print_buffer[pos++] = ' ';
-	if (album) {
-		print_buffer[pos++] = ' ';
-		print_buffer[pos++] = ' ';
-		str = album->name;
-	} else {
-		if (display_artist_sort_name)
-			str = artist_sort_name(artist);
-		else
-			str = artist->name;
-	}
-	pos += format_str(print_buffer + pos, str, tree_win_w - pos - 1);
-	print_buffer[pos++] = ' ';
-	print_buffer[pos++] = 0;
-	dump_print_buffer(tree_win_y + row + 1, tree_win_x);
-}
-
 static inline void fopt_set_str(struct format_option *fopt, const char *str)
 {
 	BUG_ON(fopt->type != FO_STR);
@@ -577,25 +568,27 @@ static void fill_track_fopts_track_info(struct track_info *info)
 	if (using_utf8) {
 		filename = info->filename;
 	} else {
-		utf8_encode(info->filename);
+		utf8_encode_to_buf(info->filename);
 		filename = conv_buffer;
 	}
 
 	fopt_set_str(&track_fopts[TF_ALBUMARTIST], info->albumartist);
 	fopt_set_str(&track_fopts[TF_ARTIST], info->artist);
 	fopt_set_str(&track_fopts[TF_ALBUM], info->album);
+	fopt_set_int(&track_fopts[TF_PLAY_COUNT], info->play_count, 0);
 	fopt_set_int(&track_fopts[TF_DISC], info->discnumber, info->discnumber == -1);
 	fopt_set_int(&track_fopts[TF_TRACK], info->tracknumber, info->tracknumber == -1);
 	fopt_set_str(&track_fopts[TF_TITLE], info->title);
-	fopt_set_str(&track_fopts[TF_YEAR], keyvals_get_val(info->comments, "date"));
+	fopt_set_int(&track_fopts[TF_YEAR], info->date / 10000, info->date <= 0);
 	fopt_set_str(&track_fopts[TF_GENRE], info->genre);
 	fopt_set_str(&track_fopts[TF_COMMENT], info->comment);
 	fopt_set_time(&track_fopts[TF_DURATION], info->duration, info->duration == -1);
+	fopt_set_int(&track_fopts[TF_DURATION_SEC], info->duration, info->duration == -1);
 	fopt_set_double(&track_fopts[TF_RG_TRACK_GAIN], info->rg_track_gain, isnan(info->rg_track_gain));
 	fopt_set_double(&track_fopts[TF_RG_TRACK_PEAK], info->rg_track_peak, isnan(info->rg_track_peak));
 	fopt_set_double(&track_fopts[TF_RG_ALBUM_GAIN], info->rg_album_gain, isnan(info->rg_album_gain));
 	fopt_set_double(&track_fopts[TF_RG_ALBUM_PEAK], info->rg_album_peak, isnan(info->rg_album_peak));
-	fopt_set_str(&track_fopts[TF_ORIGINALYEAR], keyvals_get_val(info->comments, "originaldate"));
+	fopt_set_int(&track_fopts[TF_ORIGINALYEAR], info->date / 10000, info->originaldate <= 0);
 	fopt_set_int(&track_fopts[TF_BITRATE], (int) (info->bitrate / 1000. + 0.5), info->bitrate == -1);
 	fopt_set_str(&track_fopts[TF_CODEC], info->codec);
 	fopt_set_str(&track_fopts[TF_CODEC_PROFILE], info->codec_profile);
@@ -613,7 +606,8 @@ static void fill_track_fopts_track_info(struct track_info *info)
 	fopt_set_str(&track_fopts[TF_PARTNUMBER], keyvals_get_val(info->comments, "partnumber"));
 	fopt_set_str(&track_fopts[TF_PART], keyvals_get_val(info->comments, "part"));
 	fopt_set_str(&track_fopts[TF_SUBTITLE], keyvals_get_val(info->comments, "subtitle"));
-	fopt_set_str(&track_fopts[TF_MEDIA], keyvals_get_val(info->comments, "media"));
+	fopt_set_str(&track_fopts[TF_MEDIA], info->media);
+	fopt_set_int(&track_fopts[TF_VA], 0, !track_is_compilation(info->comments));
 	if (is_http_url(info->filename)) {
 		fopt_set_str(&track_fopts[TF_FILE], filename);
 	} else {
@@ -621,15 +615,152 @@ static void fill_track_fopts_track_info(struct track_info *info)
 	}
 }
 
+static void fill_track_fopts_album(struct album *album)
+{
+	fopt_set_int(&track_fopts[TF_YEAR], album->min_date / 10000, album->min_date <= 0);
+	fopt_set_int(&track_fopts[TF_MAX_YEAR], album->date / 10000, album->date <= 0);
+	fopt_set_str(&track_fopts[TF_ALBUMARTIST], album->artist->name);
+	fopt_set_str(&track_fopts[TF_ARTIST], album->artist->name);
+	fopt_set_str(&track_fopts[TF_ALBUM], album->name);
+}
+
+static void fill_track_fopts_artist(struct artist *artist)
+{
+	const char *name = display_artist_sort_name ? artist_sort_name(artist) : artist->name;
+	fopt_set_str(&track_fopts[TF_ARTIST], name);
+}
+
+const struct format_option *get_global_fopts(void)
+{
+	if (player_info.ti)
+		fill_track_fopts_track_info(player_info.ti);
+
+	static const char *status_strs[] = { ".", ">", "|" };
+	static const char *cont_strs[] = { " ", "C" };
+	static const char *follow_strs[] = { " ", "F" };
+	static const char *repeat_strs[] = { " ", "R" };
+	static const char *shuffle_strs[] = { " ", "S" };
+	int buffer_fill, vol, vol_left, vol_right;
+	int duration = -1;
+
+	editable_lock();
+	fopt_set_time(&track_fopts[TF_TOTAL], play_library ? lib_editable.total_time :
+			pl_editable.total_time, 0);
+	editable_unlock();
+
+	fopt_set_str(&track_fopts[TF_FOLLOW], follow_strs[follow]);
+	fopt_set_str(&track_fopts[TF_REPEAT], repeat_strs[repeat]);
+	fopt_set_str(&track_fopts[TF_SHUFFLE], shuffle_strs[shuffle]);
+	fopt_set_str(&track_fopts[TF_PLAYLISTMODE], aaa_mode_names[aaa_mode]);
+
+	if (player_info.ti)
+		duration = player_info.ti->duration;
+
+	vol_left = vol_right = vol = -1;
+	if (soft_vol) {
+		vol_left = soft_vol_l;
+		vol_right = soft_vol_r;
+		vol = (vol_left + vol_right + 1) / 2;
+	} else if (volume_max && volume_l >= 0 && volume_r >= 0) {
+		vol_left = scale_to_percentage(volume_l, volume_max);
+		vol_right = scale_to_percentage(volume_r, volume_max);
+		vol = (vol_left + vol_right + 1) / 2;
+	}
+	buffer_fill = scale_to_percentage(player_info.buffer_fill, player_info.buffer_size);
+
+	fopt_set_str(&track_fopts[TF_STATUS], status_strs[player_info.status]);
+
+	if (show_remaining_time && duration != -1) {
+		fopt_set_time(&track_fopts[TF_POSITION], player_info.pos - duration, 0);
+	} else {
+		fopt_set_time(&track_fopts[TF_POSITION], player_info.pos, 0);
+	}
+
+	fopt_set_int(&track_fopts[TF_POSITION_SEC], player_info.pos, player_info.pos < 0);
+	fopt_set_time(&track_fopts[TF_DURATION], duration, duration < 0);
+	fopt_set_int(&track_fopts[TF_VOLUME], vol, vol < 0);
+	fopt_set_int(&track_fopts[TF_LVOLUME], vol_left, vol_left < 0);
+	fopt_set_int(&track_fopts[TF_RVOLUME], vol_right, vol_right < 0);
+	fopt_set_int(&track_fopts[TF_BUFFER], buffer_fill, 0);
+	fopt_set_str(&track_fopts[TF_CONTINUE], cont_strs[player_cont]);
+	fopt_set_int(&track_fopts[TF_BITRATE], player_info.current_bitrate / 1000. + 0.5, 0);
+
+	return track_fopts;
+}
+
+static void print_tree(struct window *win, int row, struct iter *iter)
+{
+	struct artist *artist;
+	struct album *album;
+	struct iter sel;
+	int current, selected, active, pos;
+
+	artist = iter_to_artist(iter);
+	album = iter_to_album(iter);
+	current = 0;
+	if (lib_cur_track) {
+		if (album) {
+			current = CUR_ALBUM == album;
+		} else {
+			current = CUR_ARTIST == artist;
+		}
+	}
+	window_get_sel(win, &sel);
+	selected = iters_equal(iter, &sel);
+	active = lib_cur_win == lib_tree_win;
+	bkgdset(pairs[(active << 2) | (selected << 1) | current]);
+
+	if (active && selected) {
+		cursor_x = 0;
+		cursor_y = 1 + row;
+	}
+
+	print_buffer[0] = ' ';
+	if (album) {
+		fill_track_fopts_album(album);
+		format_print(print_buffer + 1, tree_win_w - 2, tree_win_format, track_fopts);
+	} else {
+		fill_track_fopts_artist(artist);
+		format_print(print_buffer + 1, tree_win_w - 2, tree_win_artist_format, track_fopts);
+	}
+	pos = strlen(print_buffer);
+	print_buffer[pos++] = ' ';
+	print_buffer[pos++] = 0;
+	dump_print_buffer(tree_win_y + row + 1, tree_win_x);
+}
+
 static void print_track(struct window *win, int row, struct iter *iter)
 {
 	struct tree_track *track;
+	struct album *album;
 	struct track_info *ti;
 	struct iter sel;
 	int current, selected, active;
 	const char *format;
 
 	track = iter_to_tree_track(iter);
+	album = iter_to_album(iter);
+
+	if (track == (struct tree_track*)album) {
+		int pos;
+		struct fp_len len;
+
+		/* FIXME:
+		 * replace A_BOLD by something useful */
+		bkgdset(A_BOLD);
+
+		fill_track_fopts_album(album);
+
+		len = format_print(print_buffer, track_win_w, track_win_album_format, track_fopts);
+		dump_print_buffer(track_win_y + row + 1, track_win_x);
+
+		bkgdset(pairs[CURSED_SEPARATOR]);
+		for(pos = track_win_x + len.llen; pos < COLS - len.rlen; ++pos)
+			(void) mvaddch(track_win_y + row + 1, pos, ACS_HLINE);
+
+		return;
+	}
+
 	current = lib_cur_track == track;
 	window_get_sel(win, &sel);
 	selected = iters_equal(iter, &sel);
@@ -644,16 +775,14 @@ static void print_track(struct window *win, int row, struct iter *iter)
 	ti = tree_track_info(track);
 	fill_track_fopts_track_info(ti);
 
+	format = track_win_format;
 	if (track_info_has_tag(ti)) {
-		if (track_is_compilation(ti->comments))
+		if (*track_win_format_va && track_is_compilation(ti->comments))
 			format = track_win_format_va;
-		else
-			format = track_win_format;
-
-		format_print(print_buffer, track_win_w, format, track_fopts);
-	} else {
-		format_print(print_buffer, track_win_w, track_win_alt_format, track_fopts);
+	} else if (*track_win_alt_format) {
+		format = track_win_alt_format;
 	}
+	format_print(print_buffer, track_win_w, format, track_fopts);
 	dump_print_buffer(track_win_y + row + 1, track_win_x);
 }
 
@@ -687,16 +816,14 @@ static void print_editable(struct window *win, int row, struct iter *iter)
 
 	fill_track_fopts_track_info(track->info);
 
+	format = list_win_format;
 	if (track_info_has_tag(track->info)) {
-		if (track_is_compilation(track->info->comments))
+		if (*list_win_format_va && track_is_compilation(track->info->comments))
 			format = list_win_format_va;
-		else
-			format = list_win_format;
-
-		format_print(print_buffer, COLS, format, track_fopts);
-	} else {
-		format_print(print_buffer, COLS, list_win_alt_format, track_fopts);
+	} else if (*list_win_alt_format) {
+		format = list_win_alt_format;
 	}
+	format_print(print_buffer, COLS, format, track_fopts);
 	dump_print_buffer(row + 1, 0);
 }
 
@@ -769,7 +896,7 @@ static void print_filter(struct window *win, int row, struct iter *iter)
 
 	e_filter = e->filter;
 	if (!using_utf8) {
-		utf8_encode(e_filter);
+		utf8_encode_to_buf(e_filter);
 		e_filter = conv_buffer;
 	}
 
@@ -786,7 +913,7 @@ static void print_help(struct window *win, int row, struct iter *iter)
 	int selected;
 	int pos;
 	int active = 1;
-	char buf[512];
+	char buf[OPTION_MAX_SIZE];
 	const struct help_entry *e = iter_to_help_entry(iter);
 	const struct cmus_opt *opt;
 
@@ -904,7 +1031,7 @@ static void update_editable_window(struct editable *e, const char *title, const
 		if (using_utf8) {
 			/* already UTF-8 */
 		} else {
-			utf8_encode(filename);
+			utf8_encode_to_buf(filename);
 			filename = conv_buffer;
 		}
 		snprintf(buf, sizeof(buf), "%s %s - %d tracks", title,
@@ -951,7 +1078,7 @@ static void update_browser_window(void)
 		/* already UTF-8 */
 		dirname = browser_dir;
 	} else {
-		utf8_encode(browser_dir);
+		utf8_encode_to_buf(browser_dir);
 		dirname = conv_buffer;
 	}
 	snprintf(title, sizeof(title), "Browser - %s", dirname);
@@ -1025,94 +1152,10 @@ static void do_update_view(int full)
 
 static void do_update_statusline(void)
 {
-	static const char *status_strs[] = { ".", ">", "|" };
-	static const char *cont_strs[] = { " ", "C" };
-	static const char *repeat_strs[] = { " ", "R" };
-	static const char *shuffle_strs[] = { " ", "S" };
-	int buffer_fill, vol, vol_left, vol_right;
-	int duration = -1;
-	char *msg;
-	char format[80];
-
-	editable_lock();
-	fopt_set_time(&status_fopts[SF_TOTAL], play_library ? lib_editable.total_time :
-			pl_editable.total_time, 0);
-	editable_unlock();
-
-	fopt_set_str(&status_fopts[SF_REPEAT], repeat_strs[repeat]);
-	fopt_set_str(&status_fopts[SF_SHUFFLE], shuffle_strs[shuffle]);
-	fopt_set_str(&status_fopts[SF_PLAYLISTMODE], aaa_mode_names[aaa_mode]);
-
 	player_info_lock();
+	format_print(print_buffer, COLS, statusline_format, get_global_fopts());
 
-	if (player_info.ti)
-		duration = player_info.ti->duration;
-
-	vol_left = vol_right = vol = -1;
-	if (soft_vol) {
-		vol_left = soft_vol_l;
-		vol_right = soft_vol_r;
-		vol = (vol_left + vol_right + 1) / 2;
-	} else if (volume_max && volume_l >= 0 && volume_r >= 0) {
-		vol_left = scale_to_percentage(volume_l, volume_max);
-		vol_right = scale_to_percentage(volume_r, volume_max);
-		vol = (vol_left + vol_right + 1) / 2;
-	}
-	buffer_fill = scale_to_percentage(player_info.buffer_fill, player_info.buffer_size);
-
-	fopt_set_str(&status_fopts[SF_STATUS], status_strs[player_info.status]);
-
-	if (show_remaining_time && duration != -1) {
-		fopt_set_time(&status_fopts[SF_POSITION], player_info.pos - duration, 0);
-	} else {
-		fopt_set_time(&status_fopts[SF_POSITION], player_info.pos, 0);
-	}
-
-	fopt_set_time(&status_fopts[SF_DURATION], duration, 0);
-	fopt_set_int(&status_fopts[SF_VOLUME], vol, 0);
-	fopt_set_int(&status_fopts[SF_LVOLUME], vol_left, 0);
-	fopt_set_int(&status_fopts[SF_RVOLUME], vol_right, 0);
-	fopt_set_int(&status_fopts[SF_BUFFER], buffer_fill, 0);
-	fopt_set_str(&status_fopts[SF_CONTINUE], cont_strs[player_cont]);
-	fopt_set_int(&status_fopts[SF_BITRATE], player_info.current_bitrate / 1000. + 0.5, 0);
-
-	strcpy(format, " %s %p ");
-	if (duration != -1)
-		strcat(format, "/ %d ");
-	strcat(format, "- %t ");
-	if (vol >= 0) {
-		if (vol_left != vol_right) {
-			strcat(format, "vol: %l,%r ");
-		} else {
-			strcat(format, "vol: %v ");
-		}
-	}
-	if (player_info.ti) {
-		if (is_http_url(player_info.ti->filename))
-			strcat(format, "buf: %b ");
-		if (show_current_bitrate && player_info.current_bitrate >= 0)
-			strcat(format, " %B kbps ");
-	}
-	strcat(format, "%=");
-	if (player_repeat_current) {
-		strcat(format, "repeat current");
-	} else if (play_library) {
-		/* artist/album modes work only in lib */
-		if (shuffle) {
-			/* shuffle overrides sorted mode */
-			strcat(format, "%L from library");
-		} else if (play_sorted) {
-			strcat(format, "%L from sorted library");
-		} else {
-			strcat(format, "%L from library");
-		}
-	} else {
-		strcat(format, "playlist");
-	}
-	strcat(format, " | %1C%1R%1S ");
-	format_print(print_buffer, COLS, format, status_fopts);
-
-	msg = player_info.error_msg;
+	char *msg = player_info.error_msg;
 	player_info.error_msg = NULL;
 
 	player_info_unlock();
@@ -1174,7 +1217,7 @@ static void do_update_commandline(void)
 		 * (displayed as <xx>) there would be no problem because bpos
 		 * still equals to cpos, I think.
 		 */
-		utf8_encode(cmdline.line);
+		utf8_encode_to_buf(cmdline.line);
 		str = conv_buffer;
 	}
 
@@ -1235,28 +1278,6 @@ static void do_update_commandline(void)
 	}
 }
 
-/* lock player_info! */
-static const char *get_stream_title(void)
-{
-	static char stream_title[255 * 16 + 1];
-	char *ptr, *title;
-
-	ptr = strstr(player_info.metadata, "StreamTitle='");
-	if (ptr == NULL)
-		return NULL;
-	ptr += 13;
-	title = ptr;
-	while (*ptr) {
-		if (*ptr == '\'' && *(ptr + 1) == ';') {
-			memcpy(stream_title, title, ptr - title);
-			stream_title[ptr - title] = 0;
-			return stream_title;
-		}
-		ptr++;
-	}
-	return NULL;
-}
-
 static void set_title(const char *title)
 {
 	if (!set_term_title)
@@ -1284,15 +1305,17 @@ static void do_update_titleline(void)
 			const char *title = get_stream_title();
 
 			if (title != NULL) {
+				free(title_buf);
+				title_buf = to_utf8(title, icecast_default_charset);
 				/*
 				 * StreamTitle overrides radio station name
 				 */
 				use_alt_format = 0;
-				fopt_set_str(&track_fopts[TF_TITLE], title);
+				fopt_set_str(&track_fopts[TF_TITLE], title_buf);
 			}
 		}
 
-		if (use_alt_format) {
+		if (use_alt_format && *current_alt_format) {
 			format_print(print_buffer, COLS, current_alt_format, track_fopts);
 		} else {
 			format_print(print_buffer, COLS, current_format, track_fopts);
@@ -1300,7 +1323,7 @@ static void do_update_titleline(void)
 		dump_print_buffer(LINES - 3, 0);
 
 		/* set window title */
-		if (use_alt_format) {
+		if (use_alt_format && *window_title_alt_format) {
 			format_print(print_buffer, print_buffer_max_width,
 					window_title_alt_format, track_fopts);
 		} else {
@@ -1339,7 +1362,7 @@ static int cmdline_cursor_column(void)
 	str = cmdline.line;
 	if (!using_utf8) {
 		/* see do_update_commandline */
-		utf8_encode(cmdline.line);
+		utf8_encode_to_buf(cmdline.line);
 		str = conv_buffer;
 	}
 
@@ -1396,6 +1419,28 @@ static void post_update(void)
 	}
 }
 
+/* lock player_info! */
+const char *get_stream_title(void)
+{
+	static char stream_title[255 * 16 + 1];
+	char *ptr, *title;
+
+	ptr = strstr(player_info.metadata, "StreamTitle='");
+	if (ptr == NULL)
+		return NULL;
+	ptr += 13;
+	title = ptr;
+	while (*ptr) {
+		if (*ptr == '\'' && *(ptr + 1) == ';') {
+			memcpy(stream_title, title, ptr - title);
+			stream_title[ptr - title] = 0;
+			return stream_title;
+		}
+		ptr++;
+	}
+	return NULL;
+}
+
 void update_titleline(void)
 {
 	curs_set(0);
@@ -1657,15 +1702,16 @@ void update_colors(void)
 	for (i = 0; i < NR_CURSED; i++) {
 		int bg = colors[cursed_to_bg_idx[i]];
 		int fg = colors[cursed_to_fg_idx[i]];
+		int attr = attrs[cursed_to_attr_idx[i]];
 		int pair = i + 1;
 
 		if (fg >= 8 && fg <= 15) {
 			/* fg colors 8..15 are special (0..7 + bold) */
 			init_pair(pair, fg & 7, bg);
-			pairs[i] = COLOR_PAIR(pair) | (fg & BRIGHT ? A_BOLD : 0);
+			pairs[i] = COLOR_PAIR(pair) | (fg & BRIGHT ? A_BOLD : 0) | attr;
 		} else {
 			init_pair(pair, fg, bg);
-			pairs[i] = COLOR_PAIR(pair);
+			pairs[i] = COLOR_PAIR(pair) | attr;
 		}
 	}
 }
@@ -1759,19 +1805,19 @@ static void spawn_status_program(void)
 		free(argv[i]);
 }
 
-static int ctrl_c_pressed = 0;
+static sig_atomic_t ctrl_c_pressed = 0;
 
 static void sig_int(int sig)
 {
 	ctrl_c_pressed = 1;
 }
 
-static void sig_hup(int sig)
+static void sig_shutdown(int sig)
 {
 	cmus_running = 0;
 }
 
-static int needs_to_resize = 1;
+static sig_atomic_t needs_to_resize = 1;
 
 static void sig_winch(int sig)
 {
@@ -1898,7 +1944,7 @@ static void update(void)
 		lib_editable.win->changed = 0;
 	} else {
 		needs_status_update += pl_editable.win->changed;
-		lib_editable.win->changed = 0;
+		pl_editable.win->changed = 0;
 	}
 
 	editable_unlock();
@@ -1964,6 +2010,25 @@ static void handle_key(int key)
 	}
 }
 
+static void handle_mouse(MEVENT *event)
+{
+#if NCURSES_MOUSE_VERSION <= 1
+	if ((last_mevent & BUTTON1_PRESSED) && (event->bstate & REPORT_MOUSE_POSITION))
+		event->bstate = BUTTON1_RELEASED;
+	last_mevent = event->bstate;
+#endif
+	clear_error();
+	if (input_mode == NORMAL_MODE) {
+		normal_mode_mouse(event);
+	} else if (input_mode == COMMAND_MODE) {
+		command_mode_mouse(event);
+		update_commandline();
+	} else if (input_mode == SEARCH_MODE) {
+		search_mode_mouse(event);
+		update_commandline();
+	}
+}
+
 static void u_getch(void)
 {
 	int key;
@@ -1975,6 +2040,13 @@ static void u_getch(void)
 	if (key == ERR || key == 0)
 		return;
 
+	if (key == KEY_MOUSE) {
+		MEVENT event;
+		if (getmouse(&event) == OK)
+			handle_mouse(&event);
+		return;
+	}
+
 	if (key > 255) {
 		handle_key(key);
 		return;
@@ -2016,6 +2088,14 @@ static void u_getch(void)
 	handle_ch(u);
 }
 
+void ui_curses_notify(void)
+{
+	char c;
+	if (write(notify_in, &c, 1) == -1) {
+		d_print("write failed: %s\n", strerror(errno));
+	}
+}
+
 static void main_loop(void)
 {
 	int rc, fd_high;
@@ -2057,6 +2137,9 @@ static void main_loop(void)
 
 		FD_ZERO(&set);
 		FD_SET(0, &set);
+		FD_SET(notify_out, &set);
+		if (notify_out > fd_high)
+			fd_high = notify_out;
 		FD_SET(server_socket, &set);
 		list_for_each_entry(client, &client_head, node) {
 			FD_SET(client->fd, &set);
@@ -2123,6 +2206,13 @@ static void main_loop(void)
 
 		if (FD_ISSET(0, &set))
 			u_getch();
+
+		if (FD_ISSET(notify_out, &set)) {
+			char buf[128];
+			if (read(notify_out, buf, sizeof(buf)) == -1) {
+				d_print("read failed: %s\n", strerror(errno));
+			}
+		}
 	}
 }
 
@@ -2134,9 +2224,9 @@ static int get_next(struct track_info **ti)
 	info = play_queue_remove();
 	if (info == NULL) {
 		if (play_library) {
-			info = lib_set_next();
+			info = lib_goto_next();
 		} else {
-			info = pl_set_next();
+			info = pl_goto_next();
 		}
 	}
 	editable_unlock();
@@ -2164,8 +2254,9 @@ static void init_curses(void)
 
 	sigemptyset(&act.sa_mask);
 	act.sa_flags = 0;
-	act.sa_handler = sig_hup;
+	act.sa_handler = sig_shutdown;
 	sigaction(SIGHUP, &act, NULL);
+	sigaction(SIGTERM, &act, NULL);
 
 	sigemptyset(&act.sa_mask);
 	act.sa_flags = 0;
@@ -2236,6 +2327,17 @@ static void init_curses(void)
 			t_fs = "\007";
 		}
 	}
+	update_mouse();
+}
+
+static void notify_init(void)
+{
+	int fildes[2];
+	BUG_ON(pipe(fildes) == -1);
+	notify_out = fildes[0];
+	notify_in = fildes[1];
+	int flags = fcntl(notify_in, F_GETFL, 0);
+	fcntl(notify_in, F_SETFL, flags | O_NONBLOCK);
 }
 
 static void init_all(void)
@@ -2288,11 +2390,11 @@ static void init_all(void)
 
 	if (resume_cmus) {
 		resume_load();
-		cmus_add(play_queue_append, play_queue_autosave_filename, FILE_TYPE_PL, JOB_TYPE_QUEUE);
+		cmus_add(play_queue_append, play_queue_autosave_filename, FILE_TYPE_PL, JOB_TYPE_QUEUE, 0);
 	}
 
-	cmus_add(pl_add_track, pl_autosave_filename, FILE_TYPE_PL, JOB_TYPE_PL);
-	cmus_add(lib_add_track, lib_autosave_filename, FILE_TYPE_PL, JOB_TYPE_LIB);
+	cmus_add(pl_add_track, pl_autosave_filename, FILE_TYPE_PL, JOB_TYPE_PL, 0);
+	cmus_add(lib_add_track, lib_autosave_filename, FILE_TYPE_PL, JOB_TYPE_LIB, 0);
 }
 
 static void exit_all(void)
@@ -2341,7 +2443,7 @@ static const char *usage =
 "Usage: %s [OPTION]...\n"
 "Curses based music player.\n"
 "\n"
-"      --listen ADDR   listen on ADDR instead of ~/.cmus/socket\n"
+"      --listen ADDR   listen on ADDR instead of $CMUS_SOCKET or $XDG_RUNTIME_DIR/cmus-socket\n"
 "                      ADDR is either a UNIX socket or host[:port]\n"
 "                      WARNING: using TCP/IP is insecure!\n"
 "      --plugins       list available plugins and exit\n"
@@ -2373,7 +2475,7 @@ int main(int argc, char *argv[])
 		case FLAG_VERSION:
 			printf("cmus " VERSION
 			       "\nCopyright 2004-2006 Timo Hirvonen"
-			       "\nCopyright 2008-2011 Various Authors\n");
+			       "\nCopyright 2008-2013 Various Authors\n");
 			return 0;
 		case FLAG_PLUGINS:
 			list_plugins = 1;
@@ -2402,7 +2504,7 @@ int main(int argc, char *argv[])
 
 	misc_init();
 	if (server_address == NULL)
-		server_address = xstrjoin(cmus_config_dir, "/socket");
+		server_address = xstrdup(cmus_socket_path);
 	debug_init();
 	d_print("charset = '%s'\n", charset);
 
@@ -2413,6 +2515,7 @@ int main(int argc, char *argv[])
 		op_dump_plugins();
 		return 0;
 	}
+	notify_init();
 	init_all();
 	main_loop();
 	exit_all();
