@@ -1,5 +1,5 @@
 /*
- * Copyright 2008-2011 Various Authors
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
  *
  * This program is free software; you can redistribute it and/or
@@ -16,8 +16,8 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
-#ifndef _WINDOW_H
-#define _WINDOW_H
+#ifndef CMUS_WINDOW_H
+#define CMUS_WINDOW_H
 
 #include "iter.h"
 
@@ -54,6 +54,7 @@ struct window {
 	/* return 1 if got next/prev, otherwise 0 */
 	int (*get_prev)(struct iter *iter);
 	int (*get_next)(struct iter *iter);
+	int (*selectable)(struct iter *iter);
 	void (*sel_changed)(void);
 };
 
@@ -89,6 +90,11 @@ void window_goto_top(struct window *win);
 void window_goto_bottom(struct window *win);
 void window_page_up(struct window *win);
 void window_page_down(struct window *win);
+void window_scroll_down(struct window *win);
+void window_scroll_up(struct window *win);
+void window_page_top(struct window *win);
+void window_page_bottom(struct window *win);
+void window_page_middle(struct window *win);
 
 int window_get_nr_rows(struct window *win);
 
